###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

"""
    Ideal NaOCl Chlorination posttreatment process

    This will build an ideal NaOCl pretreatment block as a combination of a
    Mixer (where NaOCl is added) and an EquilibriumReactor (where pH and free
    chlorine is calculated)::

                    NaOCl stream
                        |
                        V
        inlet stream ---> [Mixer] --- outlet stream ---> [EquilibriumReactor] ---> exit stream (to distribution)
"""

# Importing the object for units from pyomo
from pyomo.environ import units as pyunits

# Imports from idaes core
from idaes.core import AqueousPhase
from idaes.core.components import Solvent, Solute, Cation, Anion
from idaes.core.phases import PhaseType as PT

# Imports from idaes generic models
import idaes.generic_models.properties.core.pure.Perrys as Perrys
from idaes.generic_models.properties.core.state_definitions import FTPx
from idaes.generic_models.properties.core.eos.ideal import Ideal

# Importing the enum for concentration unit basis used in the 'get_concentration_term' function
from idaes.generic_models.properties.core.generic.generic_reaction import ConcentrationForm

# Import the object/function for heat of reaction
from idaes.generic_models.properties.core.reactions.dh_rxn import constant_dh_rxn

# Import safe log power law equation
from idaes.generic_models.properties.core.reactions.equilibrium_forms import log_power_law_equil

# Import built-in van't Hoff function
from idaes.generic_models.properties.core.reactions.equilibrium_constant import van_t_hoff

# Import specific pyomo objects
from pyomo.environ import (ConcreteModel,
                           Var,
                           Constraint,
                           SolverStatus,
                           TerminationCondition,
                           TransformationFactory,
                           value,
                           Suffix,
                           Expression)

from pyomo.network import Arc

from idaes.core.util import scaling as iscale
from idaes.core.util.initialization import fix_state_vars, revert_state_vars

# Import pyomo methods to check the system units
from pyomo.util.check_units import assert_units_consistent


from watertap.examples.flowsheets.full_treatment_train.util import solve_with_user_scaling, check_dof
from watertap.examples.flowsheets.full_treatment_train.model_components import property_models
from idaes.core.util import get_solver

# Import the idaes objects for Generic Properties and Reactions
from idaes.generic_models.properties.core.generic.generic_property import (
        GenericParameterBlock)
from idaes.generic_models.properties.core.generic.generic_reaction import (
        GenericReactionParameterBlock)

# Import the idaes object for the EquilibriumReactor unit model
from idaes.generic_models.unit_models.equilibrium_reactor import EquilibriumReactor

# Import the WaterTAP object inherited for the Mixer unit model
from watertap.examples.flowsheets.full_treatment_train.model_components import Mixer

# Import costing and financials to test
from watertap.examples.flowsheets.full_treatment_train.flowsheet_components import costing, financials

from idaes.generic_models.unit_models.translator import Translator

# Import the core idaes objects for Flowsheets and types of balances
from idaes.core import FlowsheetBlock

# Import log10 function from pyomo
from pyomo.environ import log10

import idaes.logger as idaeslog

# Grab the scaling utilities
from watertap.examples.flowsheets.full_treatment_train.electrolyte_scaling_utils import (
    approximate_chemical_state_args,
    calculate_chemical_scaling_factors,
    calculate_chemical_scaling_factors_for_energy_balances)

from watertap.examples.flowsheets.full_treatment_train.chemical_flowsheet_util import (
    set_H2O_molefraction, zero_out_non_H2O_molefractions, fix_all_molefractions,
    unfix_all_molefractions, seq_decomp_initializer )

__author__ = "Austin Ladshaw"

# Configuration dictionary
ideal_naocl_thermo_config = {
    "components": {
        'H2O': {"type": Solvent,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Perrys,
              "enth_mol_liq_comp": Perrys,
              "cp_mol_liq_comp": Perrys,
              "entr_mol_liq_comp": Perrys,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (18.0153, pyunits.g/pyunits.mol),
                    "pressure_crit": (220.64E5, pyunits.Pa),
                    "temperature_crit": (647, pyunits.K),
                    # Comes from Perry's Handbook:  p. 2-98
                    "dens_mol_liq_comp_coeff": {
                        '1': (5.459, pyunits.kmol*pyunits.m**-3),
                        '2': (0.30542, pyunits.dimensionless),
                        '3': (647.13, pyunits.K),
                        '4': (0.081, pyunits.dimensionless)},
                    "enth_mol_form_liq_comp_ref": (-285.830, pyunits.kJ/pyunits.mol),
                    "enth_mol_form_vap_comp_ref": (0, pyunits.kJ/pyunits.mol),
                    # Comes from Perry's Handbook:  p. 2-174
                    "cp_mol_liq_comp_coeff": {
                        '1': (2.7637E5, pyunits.J/pyunits.kmol/pyunits.K),
                        '2': (-2.0901E3, pyunits.J/pyunits.kmol/pyunits.K**2),
                        '3': (8.125, pyunits.J/pyunits.kmol/pyunits.K**3),
                        '4': (-1.4116E-2, pyunits.J/pyunits.kmol/pyunits.K**4),
                        '5': (9.3701E-6, pyunits.J/pyunits.kmol/pyunits.K**5)},
                    "cp_mol_ig_comp_coeff": {
                        'A': (30.09200, pyunits.J/pyunits.mol/pyunits.K),
                        'B': (6.832514, pyunits.J*pyunits.mol**-1*pyunits.K**-1*pyunits.kiloK**-1),
                        'C': (6.793435, pyunits.J*pyunits.mol**-1*pyunits.K**-1*pyunits.kiloK**-2),
                        'D': (-2.534480, pyunits.J*pyunits.mol**-1*pyunits.K**-1*pyunits.kiloK**-3),
                        'E': (0.082139, pyunits.J*pyunits.mol**-1*pyunits.K**-1*pyunits.kiloK**2),
                        'F': (-250.8810, pyunits.kJ/pyunits.mol),
                        'G': (223.3967, pyunits.J/pyunits.mol/pyunits.K),
                        'H': (0, pyunits.kJ/pyunits.mol)},
                    "entr_mol_form_liq_comp_ref": (69.95, pyunits.J/pyunits.K/pyunits.mol),
                    "pressure_sat_comp_coeff": {
                        'A': (4.6543, None),  # [1], temperature range 255.9 K - 373 K
                        'B': (1435.264, pyunits.K),
                        'C': (-64.848, pyunits.K)}
                                },
                    # End parameter_data
                    },
        'H_+': {"type": Cation, "charge": 1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Perrys,
              "enth_mol_liq_comp": Perrys,
              "cp_mol_liq_comp": Perrys,
              "entr_mol_liq_comp": Perrys,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (1.00784, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": {
                        '1': (5.459, pyunits.kmol*pyunits.m**-3),
                        '2': (0.30542, pyunits.dimensionless),
                        '3': (647.13, pyunits.K),
                        '4': (0.081, pyunits.dimensionless)},
                    "enth_mol_form_liq_comp_ref": (-230.000, pyunits.kJ/pyunits.mol),
                    "cp_mol_liq_comp_coeff": {
                        '1': (2.7637E5, pyunits.J/pyunits.kmol/pyunits.K),
                        '2': (-2.0901E3, pyunits.J/pyunits.kmol/pyunits.K**2),
                        '3': (8.125, pyunits.J/pyunits.kmol/pyunits.K**3),
                        '4': (-1.4116E-2, pyunits.J/pyunits.kmol/pyunits.K**4),
                        '5': (9.3701E-6, pyunits.J/pyunits.kmol/pyunits.K**5)},
                    "entr_mol_form_liq_comp_ref": (-10.75, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'Na_+': {"type": Cation, "charge": 1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Perrys,
              "enth_mol_liq_comp": Perrys,
              "cp_mol_liq_comp": Perrys,
              "entr_mol_liq_comp": Perrys,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (22.989769, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": {
                        '1': (5.252, pyunits.kmol*pyunits.m**-3),
                        '2': (0.347, pyunits.dimensionless),
                        '3': (1595.8, pyunits.K),
                        '4': (0.6598, pyunits.dimensionless)},
                    "enth_mol_form_liq_comp_ref": (-240.1, pyunits.J/pyunits.mol),
                    "cp_mol_liq_comp_coeff": {
                        '1': (167039, pyunits.J/pyunits.kmol/pyunits.K),
                        '2': (0, pyunits.J/pyunits.kmol/pyunits.K**2),
                        '3': (0, pyunits.J/pyunits.kmol/pyunits.K**3),
                        '4': (0, pyunits.J/pyunits.kmol/pyunits.K**4),
                        '5': (0, pyunits.J/pyunits.kmol/pyunits.K**5)},
                    "entr_mol_form_liq_comp_ref": (59, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'Cl_-': {"type": Anion, "charge": -1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Perrys,
              "enth_mol_liq_comp": Perrys,
              "cp_mol_liq_comp": Perrys,
              "entr_mol_liq_comp": Perrys,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (35.453, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": {
                        '1': (4.985, pyunits.kmol*pyunits.m**-3),
                        '2': (0.36, pyunits.dimensionless),
                        '3': (1464.06, pyunits.K),
                        '4': (0.739, pyunits.dimensionless)},
                    "enth_mol_form_liq_comp_ref": (-167.2, pyunits.kJ/pyunits.mol),
                    "cp_mol_liq_comp_coeff": {
                        '1': (83993.8, pyunits.J/pyunits.kmol/pyunits.K),
                        '2': (0, pyunits.J/pyunits.kmol/pyunits.K**2),
                        '3': (0, pyunits.J/pyunits.kmol/pyunits.K**3),
                        '4': (0, pyunits.J/pyunits.kmol/pyunits.K**4),
                        '5': (0, pyunits.J/pyunits.kmol/pyunits.K**5)},
                    "entr_mol_form_liq_comp_ref": (56.5, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'OH_-': {"type": Anion,
                "charge": -1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Perrys,
              "enth_mol_liq_comp": Perrys,
              "cp_mol_liq_comp": Perrys,
              "entr_mol_liq_comp": Perrys,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (17.008, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": {
                        '1': (5.459, pyunits.kmol*pyunits.m**-3),
                        '2': (0.30542, pyunits.dimensionless),
                        '3': (647.13, pyunits.K),
                        '4': (0.081, pyunits.dimensionless)},
                    "enth_mol_form_liq_comp_ref": (-230.000, pyunits.kJ/pyunits.mol),
                    "cp_mol_liq_comp_coeff": {
                        '1': (2.7637E5, pyunits.J/pyunits.kmol/pyunits.K),
                        '2': (-2.0901E3, pyunits.J/pyunits.kmol/pyunits.K**2),
                        '3': (8.125, pyunits.J/pyunits.kmol/pyunits.K**3),
                        '4': (-1.4116E-2, pyunits.J/pyunits.kmol/pyunits.K**4),
                        '5': (9.3701E-6, pyunits.J/pyunits.kmol/pyunits.K**5)},
                    "entr_mol_form_liq_comp_ref": (-10.75, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'HOCl': {"type": Solute, "valid_phase_types": PT.aqueousPhase,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Perrys,
              "enth_mol_liq_comp": Perrys,
              "cp_mol_liq_comp": Perrys,
              "entr_mol_liq_comp": Perrys,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (52.46, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": {
                        '1': (4.985, pyunits.kmol*pyunits.m**-3),
                        '2': (0.36, pyunits.dimensionless),
                        '3': (1464.06, pyunits.K),
                        '4': (0.739, pyunits.dimensionless)},
                    "enth_mol_form_liq_comp_ref": (-120.9, pyunits.kJ/pyunits.mol),
                    "cp_mol_liq_comp_coeff": {
                        '1': (83993.8, pyunits.J/pyunits.kmol/pyunits.K),
                        '2': (0, pyunits.J/pyunits.kmol/pyunits.K**2),
                        '3': (0, pyunits.J/pyunits.kmol/pyunits.K**3),
                        '4': (0, pyunits.J/pyunits.kmol/pyunits.K**4),
                        '5': (0, pyunits.J/pyunits.kmol/pyunits.K**5)},
                    "entr_mol_form_liq_comp_ref": (142, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'OCl_-': {"type": Anion, "charge": -1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Perrys,
              "enth_mol_liq_comp": Perrys,
              "cp_mol_liq_comp": Perrys,
              "entr_mol_liq_comp": Perrys,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (51.46, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": {
                        '1': (4.985, pyunits.kmol*pyunits.m**-3),
                        '2': (0.36, pyunits.dimensionless),
                        '3': (1464.06, pyunits.K),
                        '4': (0.739, pyunits.dimensionless)},
                    "enth_mol_form_liq_comp_ref": (-107.1, pyunits.kJ/pyunits.mol),
                    "cp_mol_liq_comp_coeff": {
                        '1': (83993.8, pyunits.J/pyunits.kmol/pyunits.K),
                        '2': (0, pyunits.J/pyunits.kmol/pyunits.K**2),
                        '3': (0, pyunits.J/pyunits.kmol/pyunits.K**3),
                        '4': (0, pyunits.J/pyunits.kmol/pyunits.K**4),
                        '5': (0, pyunits.J/pyunits.kmol/pyunits.K**5)},
                    "entr_mol_form_liq_comp_ref": (42, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    }
              },
              # End Component list
        "phases":  {'Liq': {"type": AqueousPhase,
                            "equation_of_state": Ideal},
                    },

        "state_definition": FTPx,
        "state_bounds": {"flow_mol": (0, 50, 100),
                         "temperature": (273.15, 300, 650),
                         "pressure": (5e4, 1e5, 1e6)
                        },

        "pressure_ref": 1e5,
        "temperature_ref": 300,
        "base_units": {"time": pyunits.s,
                       "length": pyunits.m,
                       "mass": pyunits.kg,
                       "amount": pyunits.mol,
                       "temperature": pyunits.K},

    }
    # End simple_naocl_thermo_config definition

# This config is REQUIRED to use EquilibriumReactor even if we have no equilibrium reactions
ideal_naocl_reaction_config = {
    "base_units": {"time": pyunits.s,
                   "length": pyunits.m,
                   "mass": pyunits.kg,
                   "amount": pyunits.mol,
                   "temperature": pyunits.K},
    "equilibrium_reactions": {
        "H2O_Kw": {
                "stoichiometry": {("Liq", "H2O"): -1,
                                 ("Liq", "H_+"): 1,
                                 ("Liq", "OH_-"): 1},
               "heat_of_reaction": constant_dh_rxn,
               "equilibrium_constant": van_t_hoff,
               "equilibrium_form": log_power_law_equil,
               "concentration_form": ConcentrationForm.moleFraction,
               "parameter_data": {
                   "dh_rxn_ref": (55.830, pyunits.J/pyunits.mol),
                   "k_eq_ref": (10**-14/55.2/55.2, pyunits.dimensionless),
                   "T_eq_ref": (298, pyunits.K),

                   # By default, reaction orders follow stoichiometry
                   #    manually set reaction order here to override
                   "reaction_order": {("Liq", "H2O"): 0,
                                    ("Liq", "H_+"): 1,
                                    ("Liq", "OH_-"): 1}
                    }
                    # End parameter_data
               },
        "HOCl_Ka": {
                "stoichiometry": {("Liq", "HOCl"): -1,
                                 ("Liq", "H_+"): 1,
                                 ("Liq", "OCl_-"): 1},
               "heat_of_reaction": constant_dh_rxn,
               "equilibrium_constant": van_t_hoff,
               "equilibrium_form": log_power_law_equil,
               "concentration_form": ConcentrationForm.moleFraction,
               "parameter_data": {
                   "dh_rxn_ref": (13.8, pyunits.J/pyunits.mol),
                   "k_eq_ref": (10**-7.6422/55.2, pyunits.dimensionless),
                   "T_eq_ref": (298, pyunits.K),

                   # By default, reaction orders follow stoichiometry
                   #    manually set reaction order here to override
                   "reaction_order": {("Liq", "HOCl"): -1,
                                    ("Liq", "H_+"): 1,
                                    ("Liq", "OCl_-"): 1}
                    }
                    # End parameter_data
               }
               # End R2
         }
         # End equilibrium_reactions
    }
    # End reaction_config definition

# Get default solver for testing
solver = get_solver()

def build_ideal_naocl_prop(model):
    model.fs.ideal_naocl_thermo_params = GenericParameterBlock(default=ideal_naocl_thermo_config)
    model.fs.ideal_naocl_rxn_params = GenericReactionParameterBlock(
            default={"property_package": model.fs.ideal_naocl_thermo_params, **ideal_naocl_reaction_config})

def build_ideal_naocl_mixer_unit(model):
    model.fs.ideal_naocl_mixer_unit = Mixer(default={
            "property_package": model.fs.ideal_naocl_thermo_params,
            "inlet_list": ["inlet_stream", "naocl_stream"]})

    # add new constraint for dosing rate (deactivate constraint for OCl_-)
    dr = model.fs.ideal_naocl_mixer_unit.naocl_stream.flow_mol[0].value* \
            model.fs.ideal_naocl_mixer_unit.naocl_stream.mole_frac_comp[0, "OCl_-"].value
    dr = dr*74.44*1000
    model.fs.ideal_naocl_mixer_unit.dosing_rate = Var(initialize=dr)

    def _dosing_rate_cons(blk):
        return blk.dosing_rate == blk.naocl_stream.flow_mol[0]*blk.naocl_stream.mole_frac_comp[0, "OCl_-"]*74.44*1000

    model.fs.ideal_naocl_mixer_unit.dosing_cons = Constraint( rule=_dosing_rate_cons )

def build_ideal_naocl_chlorination_unit(model):
    model.fs.ideal_naocl_chlorination_unit = EquilibriumReactor(default={
            "property_package": model.fs.ideal_naocl_thermo_params,
            "reaction_package": model.fs.ideal_naocl_rxn_params,
            "has_rate_reactions": False,
            "has_equilibrium_reactions": True,
            "has_heat_transfer": False,
            "has_heat_of_reaction": False,
            "has_pressure_change": False})

    model.fs.ideal_naocl_chlorination_unit.control_volume.properties_out[0.0].log_mole_frac_phase_comp_true.setlb(-50)
    model.fs.ideal_naocl_chlorination_unit.control_volume.properties_out[0.0].log_mole_frac_phase_comp_true.setub(0.001)
    model.fs.ideal_naocl_chlorination_unit.control_volume.properties_out[0.0].mole_frac_phase_comp.setub(1.001)

    # new var includes an initial calculation (will be overwritten later)
    fc = model.fs.ideal_naocl_chlorination_unit.inlet.mole_frac_comp[0, "HOCl"].value*55.6
    fc += model.fs.ideal_naocl_chlorination_unit.inlet.mole_frac_comp[0, "OCl_-"].value*55.6
    fc = fc*70900

    model.fs.ideal_naocl_chlorination_unit.free_chlorine = Var(initialize=fc)

    def _free_chlorine_cons(blk):
        return blk.free_chlorine == ((blk.control_volume.properties_out[0.0].conc_mol_phase_comp["Liq","HOCl"]/1000) \
                                    + (blk.control_volume.properties_out[0.0].conc_mol_phase_comp["Liq","OCl_-"]/1000)) \
                                    * 70900
    model.fs.ideal_naocl_chlorination_unit.chlorine_cons = Constraint( rule=_free_chlorine_cons )

def set_ideal_naocl_mixer_inlets(model, dosing_rate_of_NaOCl_mg_per_s = 0.4,
                                        inlet_water_density_kg_per_L = 1,
                                        inlet_temperature_K = 298,
                                        inlet_pressure_Pa = 101325,
                                        inlet_flow_mol_per_s = 10):

    #inlet stream
    model.fs.ideal_naocl_mixer_unit.inlet_stream.flow_mol[0].set_value(inlet_flow_mol_per_s)
    model.fs.ideal_naocl_mixer_unit.inlet_stream.pressure[0].set_value(inlet_pressure_Pa)
    model.fs.ideal_naocl_mixer_unit.inlet_stream.temperature[0].set_value(inlet_temperature_K)

    zero_out_non_H2O_molefractions(model.fs.ideal_naocl_mixer_unit.inlet_stream)
    set_H2O_molefraction(model.fs.ideal_naocl_mixer_unit.inlet_stream)

    #naocl stream
    model.fs.ideal_naocl_mixer_unit.naocl_stream.pressure[0].set_value(inlet_pressure_Pa)
    model.fs.ideal_naocl_mixer_unit.naocl_stream.temperature[0].set_value(inlet_temperature_K)
    # Use given dosing rate value to estimate OCl_- molefraction and flow rate for naocl stream
    zero_out_non_H2O_molefractions(model.fs.ideal_naocl_mixer_unit.naocl_stream)
    model.fs.ideal_naocl_mixer_unit.naocl_stream.mole_frac_comp[0, "OCl_-"].set_value(0.5)
    model.fs.ideal_naocl_mixer_unit.naocl_stream.mole_frac_comp[0, "Na_+"].set_value(0.5)
    set_H2O_molefraction(model.fs.ideal_naocl_mixer_unit.naocl_stream)
    flow_of_naocl = dosing_rate_of_NaOCl_mg_per_s/ \
                model.fs.ideal_naocl_mixer_unit.naocl_stream.mole_frac_comp[0, "OCl_-"].value/ \
                74.44/1000
    model.fs.ideal_naocl_mixer_unit.naocl_stream.flow_mol[0].set_value(flow_of_naocl)

    model.fs.ideal_naocl_mixer_unit.dosing_rate.set_value(dosing_rate_of_NaOCl_mg_per_s)

def set_ideal_naocl_chlorination_inlets(model, mg_per_L_NaOCl_added = 2,
                                                inlet_water_density_kg_per_L = 1,
                                                inlet_temperature_K = 298,
                                                inlet_pressure_Pa = 101325,
                                                inlet_flow_mol_per_s = 10):

    zero_out_non_H2O_molefractions(model.fs.ideal_naocl_chlorination_unit.inlet)

    total_molar_density = inlet_water_density_kg_per_L/18*1000 #mol/L

    # Free Chlorine (mg-Cl2/L) = total_chlorine_inlet (mol/L) * 70,900
    #       Assumes chlorine is added as NaOCl
    free_chlorine_added = mg_per_L_NaOCl_added/74.44/1000*70900 #mg/L as NaOCl
    total_chlorine_inlet = free_chlorine_added/70900 # mol/L
    total_molar_density+=total_chlorine_inlet

    model.fs.ideal_naocl_chlorination_unit.inlet.mole_frac_comp[0, "OCl_-"].set_value( total_chlorine_inlet/total_molar_density )
    model.fs.ideal_naocl_chlorination_unit.inlet.mole_frac_comp[0, "Na_+"].set_value( total_chlorine_inlet/total_molar_density )

    set_H2O_molefraction(model.fs.ideal_naocl_chlorination_unit.inlet)

    model.fs.ideal_naocl_chlorination_unit.inlet.pressure[0].set_value(inlet_pressure_Pa)
    model.fs.ideal_naocl_chlorination_unit.inlet.temperature[0].set_value(inlet_temperature_K)
    model.fs.ideal_naocl_chlorination_unit.inlet.flow_mol[0].set_value(inlet_flow_mol_per_s)

    model.fs.ideal_naocl_chlorination_unit.free_chlorine.set_value(mg_per_L_NaOCl_added)


def fix_ideal_naocl_mixer_inlets(model):
    model.fs.ideal_naocl_mixer_unit.inlet_stream.flow_mol[0].fix()
    model.fs.ideal_naocl_mixer_unit.inlet_stream.pressure[0].fix()
    model.fs.ideal_naocl_mixer_unit.inlet_stream.temperature[0].fix()
    fix_all_molefractions(model.fs.ideal_naocl_mixer_unit.inlet_stream)

    model.fs.ideal_naocl_mixer_unit.naocl_stream.flow_mol[0].fix()
    model.fs.ideal_naocl_mixer_unit.naocl_stream.pressure[0].fix()
    model.fs.ideal_naocl_mixer_unit.naocl_stream.temperature[0].fix()
    fix_all_molefractions(model.fs.ideal_naocl_mixer_unit.naocl_stream)

def unfix_ideal_naocl_mixer_inlet_stream(model):
    model.fs.ideal_naocl_mixer_unit.inlet_stream.flow_mol[0].unfix()
    model.fs.ideal_naocl_mixer_unit.inlet_stream.pressure[0].unfix()
    model.fs.ideal_naocl_mixer_unit.inlet_stream.temperature[0].unfix()
    unfix_all_molefractions(model.fs.ideal_naocl_mixer_unit.inlet_stream)

def unfix_ideal_naocl_mixer_naocl_stream(model):
    model.fs.ideal_naocl_mixer_unit.naocl_stream.flow_mol[0].unfix()
    model.fs.ideal_naocl_mixer_unit.naocl_stream.pressure[0].unfix()
    model.fs.ideal_naocl_mixer_unit.naocl_stream.temperature[0].unfix()
    unfix_all_molefractions(model.fs.ideal_naocl_mixer_unit.naocl_stream)

def fix_ideal_naocl_chlorination_inlets(model):
    model.fs.ideal_naocl_chlorination_unit.inlet.flow_mol[0].fix()
    model.fs.ideal_naocl_chlorination_unit.inlet.pressure[0].fix()
    model.fs.ideal_naocl_chlorination_unit.inlet.temperature[0].fix()
    fix_all_molefractions(model.fs.ideal_naocl_chlorination_unit.inlet)

def unfix_ideal_naocl_chlorination_inlets(model):
    model.fs.ideal_naocl_chlorination_unit.inlet.flow_mol[0].unfix()
    model.fs.ideal_naocl_chlorination_unit.inlet.pressure[0].unfix()
    model.fs.ideal_naocl_chlorination_unit.inlet.temperature[0].unfix()
    unfix_all_molefractions(model.fs.ideal_naocl_chlorination_unit.inlet)

# basic jacobian scaling for this unit
def scale_ideal_naocl_mixer(unit):
    iscale.constraint_autoscale_large_jac(unit)

def scale_ideal_naocl_chlorination(unit, rxn_params, thermo_params, rxn_config):
    state_args, stoich_extents = approximate_chemical_state_args(unit,
                                rxn_params, rxn_config)
    calculate_chemical_scaling_factors(unit,
                                        thermo_params,
                                        rxn_params, state_args)
    return state_args

def initialize_ideal_naocl_mixer(unit, debug_out=False):
    solver.options['bound_push'] = 1e-10
    solver.options['mu_init'] = 1e-6
    solver.options["nlp_scaling_method"] = "user-scaling"
    was_fixed = False
    if unit.naocl_stream.flow_mol[0].is_fixed() == False:
        unit.naocl_stream.flow_mol[0].fix()
        was_fixed = True
    if debug_out == True:
        unit.initialize(optarg=solver.options, outlvl=idaeslog.DEBUG)
    else:
        unit.initialize(optarg=solver.options)
    if was_fixed == True:
        unit.naocl_stream.flow_mol[0].unfix()

def initialize_ideal_naocl_chlorination(unit, state_args, debug_out=False):
    solver.options['bound_push'] = 1e-10
    solver.options['mu_init'] = 1e-6
    solver.options["nlp_scaling_method"] = "user-scaling"
    if debug_out == True:
        unit.initialize(state_args=state_args, optarg=solver.options, outlvl=idaeslog.DEBUG)
    else:
        unit.initialize(state_args=state_args, optarg=solver.options)

def setup_block_to_solve_naocl_dosing_rate(model, free_chlorine_mg_per_L = 2):
    model.fs.ideal_naocl_chlorination_unit.free_chlorine.fix(free_chlorine_mg_per_L)
    model.fs.ideal_naocl_mixer_unit.naocl_stream.flow_mol[0].unfix()

def display_results_of_ideal_naocl_mixer(unit):
    print()
    print("=========== Ideal NaOCl Mixer Results ============")
    print("Outlet Temperature:       \t" + str(unit.outlet.temperature[0].value))
    print("Outlet Pressure:          \t" + str(unit.outlet.pressure[0].value))
    print("Outlet FlowMole:          \t" + str(unit.outlet.flow_mol[0].value))
    print()
    total_molar_density = 55.6
    total_salt = value(unit.outlet.mole_frac_comp[0, "Na_+"])*total_molar_density*23
    total_salt += value(unit.outlet.mole_frac_comp[0, "Cl_-"])*total_molar_density*35.4
    psu = total_salt/(total_molar_density*18)*1000
    print("STP Salinity (PSU):           \t" + str(psu))
    print("NaOCl Dosing Rate (mg/s): \t" + str(unit.dosing_rate.value))
    print("-------------------------------------------------")
    print()

def display_results_of_chlorination_unit(unit):
    print()
    print("=========== Chlorination Results ============")
    print("Outlet Temperature:       \t" + str(unit.outlet.temperature[0].value))
    print("Outlet Pressure:          \t" + str(unit.outlet.pressure[0].value))
    print("Outlet FlowMole:          \t" + str(unit.outlet.flow_mol[0].value))
    print()
    total_molar_density = \
        value(unit.control_volume.properties_out[0.0].dens_mol_phase['Liq'])/1000
    pH = -value(log10(unit.outlet.mole_frac_comp[0, "H_+"]*total_molar_density))
    print("pH at Outlet:             \t" + str(pH))
    total_salt = value(unit.outlet.mole_frac_comp[0, "Na_+"])*total_molar_density*23
    total_salt += value(unit.outlet.mole_frac_comp[0, "Cl_-"])*total_molar_density*35.4
    psu = total_salt/(total_molar_density*18)*1000
    print("Salinity (PSU):           \t" + str(psu))
    print("Free Chlorine (mg/L):     \t" + str(unit.free_chlorine.value))
    print("\tDistribution:")
    hocl = (value(unit.control_volume.properties_out[0.0].conc_mol_phase_comp["Liq","HOCl"])/1000)/(unit.free_chlorine.value/70900)
    print("\t % HOCl: \t" + str(hocl*100))
    ocl = (value(unit.control_volume.properties_out[0.0].conc_mol_phase_comp["Liq","OCl_-"])/1000)/(unit.free_chlorine.value/70900)
    print("\t % OCl-: \t" + str(ocl*100))
    print("-------------------------------------------")
    print()

def build_ideal_naocl_chlorination_block(model, expand_arcs=False):

    # Add properties to model
    build_ideal_naocl_prop(model)

    # Add mixer to the model
    build_ideal_naocl_mixer_unit(model)

    # Add reactor to the model
    build_ideal_naocl_chlorination_unit(model)

    # Connect the mixer to the chlorination unit with arcs
    model.fs.ideal_nacol_arc_mixer_to_chlor = Arc(source=model.fs.ideal_naocl_mixer_unit.outlet,
                                                destination=model.fs.ideal_naocl_chlorination_unit.inlet)
    if expand_arcs == True:
        TransformationFactory("network.expand_arcs").apply_to(model)

# This method assumes that the flowsheet has a properties object named prop_TDS
def build_translator_from_RO_to_chlorination_block(model):
    # Translator inlet from RO and outlet goes to chlorination
    # NOTE: May need to come up with a way to set state_args for Translator for
    #       better convergence behavior. This block seems to be the trouble maker
    #       for the full solve.
    model.fs.RO_to_Chlor = Translator(
        default={"inlet_property_package": model.fs.prop_TDS,
                 "outlet_property_package": model.fs.ideal_naocl_thermo_params})

    # Add constraints to define how the translator will function
    model.fs.RO_to_Chlor.eq_equal_temperature = Constraint(
        expr=model.fs.RO_to_Chlor.inlet.temperature[0]
        == model.fs.RO_to_Chlor.outlet.temperature[0])
    model.fs.RO_to_Chlor.eq_equal_pressure = Constraint(
        expr=model.fs.RO_to_Chlor.inlet.pressure[0]
        == model.fs.RO_to_Chlor.outlet.pressure[0])

    model.fs.RO_to_Chlor.total_flow_cons = Constraint(
        expr=model.fs.RO_to_Chlor.outlet.flow_mol[0] ==
            (model.fs.RO_to_Chlor.inlet.flow_mass_phase_comp[0, 'Liq', 'H2O']/18e-3) +
            (model.fs.RO_to_Chlor.inlet.flow_mass_phase_comp[0, 'Liq', 'TDS']/58.4e-3) )

    model.fs.RO_to_Chlor.H_con = Constraint( expr=model.fs.RO_to_Chlor.outlet.mole_frac_comp[0, "H_+"] == 0 )
    model.fs.RO_to_Chlor.OH_con = Constraint( expr=model.fs.RO_to_Chlor.outlet.mole_frac_comp[0, "OH_-"] == 0 )
    model.fs.RO_to_Chlor.HOCl_con = Constraint( expr=model.fs.RO_to_Chlor.outlet.mole_frac_comp[0, "HOCl"] == 0 )
    model.fs.RO_to_Chlor.OCl_con = Constraint( expr=model.fs.RO_to_Chlor.outlet.mole_frac_comp[0, "OCl_-"] == 0 )

    model.fs.RO_to_Chlor.Cl_con = Constraint(
        expr=model.fs.RO_to_Chlor.outlet.mole_frac_comp[0, "Cl_-"] ==
            (model.fs.RO_to_Chlor.inlet.flow_mass_phase_comp[0, 'Liq', 'TDS']/58.4e-3) /
             model.fs.RO_to_Chlor.outlet.flow_mol[0] )

    model.fs.RO_to_Chlor.Na_con = Constraint(
        expr=model.fs.RO_to_Chlor.outlet.mole_frac_comp[0, "Na_+"] ==
            (model.fs.RO_to_Chlor.inlet.flow_mass_phase_comp[0, 'Liq', 'TDS']/58.4e-3) /
             model.fs.RO_to_Chlor.outlet.flow_mol[0] + model.fs.RO_to_Chlor.outlet.mole_frac_comp[0, "OCl_-"])

    model.fs.RO_to_Chlor.H2O_con = Constraint(
        expr=model.fs.RO_to_Chlor.outlet.mole_frac_comp[0, "H2O"] == 1 -
            sum(model.fs.RO_to_Chlor.outlet.mole_frac_comp[0, j] for j in ["H_+", "OH_-",
                "HOCl", "OCl_-", "Cl_-", "Na_+"]) )

    iscale.calculate_scaling_factors(model.fs.RO_to_Chlor)
    iscale.constraint_autoscale_large_jac(model.fs.RO_to_Chlor)

def run_ideal_naocl_mixer_example(fixed_dosage=False):
    model = ConcreteModel()
    model.fs = FlowsheetBlock(default={"dynamic": False})

    # Add properties to model
    build_ideal_naocl_prop(model)

    # Add mixer to the model
    build_ideal_naocl_mixer_unit(model)

    # Set some inlet values
    set_ideal_naocl_mixer_inlets(model)

    # Fix the inlets for a solve
    fix_ideal_naocl_mixer_inlets(model)

    # unfix the flow_mol for the naocl_stream and fix dosing rate (alt form)
    if fixed_dosage == True:
        model.fs.ideal_naocl_mixer_unit.naocl_stream.flow_mol[0].unfix()
        model.fs.ideal_naocl_mixer_unit.dosing_rate.fix()

    check_dof(model)

    # scale the mixer
    scale_ideal_naocl_mixer(model.fs.ideal_naocl_mixer_unit)

    # initializer mixer
    initialize_ideal_naocl_mixer(model.fs.ideal_naocl_mixer_unit)

    # solve with user scaling
    solve_with_user_scaling(model, tee=True, bound_push=1e-10, mu_init=1e-6)

    display_results_of_ideal_naocl_mixer(model.fs.ideal_naocl_mixer_unit)

    return model

def run_ideal_naocl_chlorination_example():
    model = ConcreteModel()
    model.fs = FlowsheetBlock(default={"dynamic": False})

    # add properties to model
    build_ideal_naocl_prop(model)

    # add equilibrium unit
    build_ideal_naocl_chlorination_unit(model)

    # Set some inlet values
    set_ideal_naocl_chlorination_inlets(model)

    # fix the inlets
    fix_ideal_naocl_chlorination_inlets(model)

    check_dof(model)

    # scale the chlorination unit
    state_args = scale_ideal_naocl_chlorination(model.fs.ideal_naocl_chlorination_unit,
                                    model.fs.ideal_naocl_rxn_params,
                                    model.fs.ideal_naocl_thermo_params,
                                    ideal_naocl_reaction_config)

    # initialize the unit
    initialize_ideal_naocl_chlorination(model.fs.ideal_naocl_chlorination_unit, state_args, debug_out=False)

    # solve with user scaling
    solve_with_user_scaling(model, tee=True, bound_push=1e-10, mu_init=1e-6)

    display_results_of_chlorination_unit(model.fs.ideal_naocl_chlorination_unit)

    return model

def run_chlorination_block_example(fix_free_chlorine=False):
    model = ConcreteModel()
    model.fs = FlowsheetBlock(default={"dynamic": False})

    # Build the partial flowsheet of a mixer and chlorination unit
    build_ideal_naocl_chlorination_block(model, expand_arcs=True)

    # Commented section below was implemented for quick test of chlorination costing
    # # need load factor from costing_param_block for annual_water_production
    # financials.add_costing_param_block(model.fs)
    # # annual water production
    # model.fs.annual_water_production = Expression(
    #     expr=pyunits.convert(0.85 * pyunits.m**3 / pyunits.s, to_units=pyunits.m ** 3 / pyunits.year)
    #          * model.fs.costing_param.load_factor)
    # costing.build_costing(model, module=financials)


    # set some values (using defaults for testing)
    set_ideal_naocl_mixer_inlets(model, dosing_rate_of_NaOCl_mg_per_s = 0.4,
                                            inlet_water_density_kg_per_L = 1,
                                            inlet_temperature_K = 298,
                                            inlet_pressure_Pa = 101325,
                                            inlet_flow_mol_per_s = 25)
    set_ideal_naocl_chlorination_inlets(model)

    # fix only the inlets for the mixer
    fix_ideal_naocl_mixer_inlets(model)
    # scale and initialize the mixer
    scale_ideal_naocl_mixer(model.fs.ideal_naocl_mixer_unit)
    initialize_ideal_naocl_mixer(model.fs.ideal_naocl_mixer_unit)

    # scale and initialize the chlorination unit
    state_args = scale_ideal_naocl_chlorination(model.fs.ideal_naocl_chlorination_unit,
                                    model.fs.ideal_naocl_rxn_params,
                                    model.fs.ideal_naocl_thermo_params,
                                    ideal_naocl_reaction_config)

    # initialize the unit
    initialize_ideal_naocl_chlorination(model.fs.ideal_naocl_chlorination_unit, state_args, debug_out=False)

    check_dof(model)

    # Scale the full model and call the seq_decomp_initializer
    seq_decomp_initializer(model)

    # solve with user scaling
    iscale.constraint_autoscale_large_jac(model)

    if fix_free_chlorine == True:
        setup_block_to_solve_naocl_dosing_rate(model)

    # NOTE: With new log conc form, it is better to NOT have a very low bound_push
    #           inside of your flowsheet. For initialization of the block, it is
    #           usually fine. BUT, when solving the flowsheet (wth user-scaling)
    #           it seems to cause some issues
    solve_with_user_scaling(model, tee=True, bound_push=1e-4, mu_init=1e-1)

    display_results_of_ideal_naocl_mixer(model.fs.ideal_naocl_mixer_unit)
    display_results_of_chlorination_unit(model.fs.ideal_naocl_chlorination_unit)

    return model

if __name__ == "__main__":
    model = run_chlorination_block_example(fix_free_chlorine=True)
    property_models.build_prop(model, base='TDS')
    build_translator_from_RO_to_chlorination_block(model)
    # costing.display_costing(model)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for general zero-order proeprty package
"""
import pytest

from idaes.core import (ControlVolume0DBlock,
                        EnergyBalanceType,
                        FlowsheetBlock,
                        MaterialBalanceType,
                        MaterialFlowBasis)
from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              fixed_variables_set,
                                              activated_constraints_set,
                                              unused_variables_set)
import idaes.core.util.scaling as iscale
from pyomo.environ import (ConcreteModel,
                           Expression,
                           Param,
                           Set,
                           units as pyunits,
                           Var)
from pyomo.util.check_units import (assert_units_consistent,
                                    assert_units_equivalent)

from watertap.core.zero_order_properties import \
    WaterParameterBlock, WaterStateBlock


@pytest.fixture(scope="module")
def model():
    m = ConcreteModel()

    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.water_props = WaterParameterBlock(
        default={"solute_list": ["A", "B", "C"]})

    return m


@pytest.mark.unit
def test_parameter_block(model):
    assert isinstance(model.fs.water_props.component_list, Set)
    for j in model.fs.water_props.component_list:
        assert j in ["H2O", "A", "B", "C"]
    assert isinstance(model.fs.water_props.solvent_set, Set)
    for j in model.fs.water_props.solvent_set:
        assert j in ["H2O"]
    assert isinstance(model.fs.water_props.solute_set, Set)
    for j in model.fs.water_props.solute_set:
        assert j in ["A", "B", "C"]

    assert isinstance(model.fs.water_props.phase_list, Set)
    for j in model.fs.water_props.phase_list:
        assert j in ["Liq"]

    assert model.fs.water_props._state_block_class is WaterStateBlock

    assert isinstance(model.fs.water_props.pressure_ref, Param)
    assert model.fs.water_props.pressure_ref.value == 101325

    assert isinstance(model.fs.water_props.temperature_ref, Param)
    assert model.fs.water_props.temperature_ref.value == 298.15

    assert isinstance(model.fs.water_props.cp_mass, Param)
    assert model.fs.water_props.cp_mass.value == 4.184e3

    assert isinstance(model.fs.water_props.dens_mass, Param)
    assert model.fs.water_props.dens_mass.value == 1000


@pytest.mark.unit
def test_build_state_block(model):
    model.fs.state = model.fs.water_props.build_state_block([0])

    assert isinstance(model.fs.state, WaterStateBlock)

    assert model.fs.state.component_list is model.fs.water_props.component_list
    assert model.fs.state[0].component_list is \
        model.fs.water_props.component_list
    assert model.fs.state.phase_list is model.fs.water_props.phase_list
    assert model.fs.state[0].phase_list is model.fs.water_props.phase_list


@pytest.mark.unit
def test_state_block_basic_attributes(model):
    assert isinstance(model.fs.state[0].flow_vol, Var)
    assert isinstance(model.fs.state[0].conc_mass_comp, Var)
    assert isinstance(model.fs.state[0].temperature, Var)
    assert isinstance(model.fs.state[0].pressure, Var)

    # All variables are stale, so DoF should be 0
    assert len(unused_variables_set(model.fs.state[0])) == 6
    assert degrees_of_freedom(model.fs.state[0]) == 0

    assert isinstance(model.fs.state[0].flow_mass_comp, Expression)
    assert isinstance(model.fs.state[0]._enth_dens_term, Expression)
    assert isinstance(model.fs.state[0]._enth_flow_term, Expression)

    for p in model.fs.state[0].phase_list:
        assert (model.fs.state[0].get_enthalpy_flow_terms(p) is
                model.fs.state[0]._enth_flow_term)
        assert (model.fs.state[0].get_energy_density_terms(p) is
                model.fs.state[0]._enth_dens_term)

        for j in model.fs.state[0].component_list:
            assert (model.fs.state[0].get_material_flow_terms(p, j) is
                    model.fs.state[0].flow_mass_comp[j])

            if j == "H2O":
                assert (model.fs.state[0].get_material_density_terms(p, j) is
                        model.fs.water_props.dens_mass)
            else:
                assert (model.fs.state[0].get_material_density_terms(p, j) is
                        model.fs.state[0].conc_mass_comp[j])

    assert (model.fs.state[0].default_material_balance_type() is
            MaterialBalanceType.componentTotal)

    assert (model.fs.state[0].default_energy_balance_type() is
            EnergyBalanceType.enthalpyTotal)

    assert (model.fs.state[0].define_state_vars() == {
        "flow_vol": model.fs.state[0].flow_vol,
        "conc_mass_comp": model.fs.state[0].conc_mass_comp,
        "temperature": model.fs.state[0].temperature,
        "pressure": model.fs.state[0].pressure})

    assert (model.fs.state[0].define_display_vars() == {
        "Volumetric Flowrate": model.fs.state[0].flow_vol,
        "Mass Concentration": model.fs.state[0].conc_mass_comp,
        "Temperature": model.fs.state[0].temperature,
        "Pressure": model.fs.state[0].pressure})

    assert (model.fs.state[0].get_material_flow_basis() is
            MaterialFlowBasis.mass)


@pytest.mark.unit
def test_state_block_other_properties(model):
    assert model.fs.state[0].cp_mass is model.fs.water_props.cp_mass
    assert model.fs.state[0].dens_mass is model.fs.water_props.dens_mass


@pytest.mark.unit
def test_state_block_scaling(model):
    # Set some new default scaling factors
    model.fs.water_props.default_scaling_factor[("conc_mass_comp", "B")] = 5e-2

    iscale.calculate_scaling_factors(model)

    model.fs.state[0].scaling_factor.display()
    assert len(model.fs.state[0].scaling_factor) == 12

    assert model.fs.state[0].scaling_factor[model.fs.state[0].flow_vol] == 1e3
    assert model.fs.state[0].scaling_factor[
        model.fs.state[0].conc_mass_comp["A"]] == 100
    assert model.fs.state[0].scaling_factor[
        model.fs.state[0].conc_mass_comp["B"]] == 5e-2
    assert model.fs.state[0].scaling_factor[
        model.fs.state[0].conc_mass_comp["C"]] == 100
    assert model.fs.state[0].scaling_factor[
        model.fs.state[0].temperature] == 1e-2
    assert model.fs.state[0].scaling_factor[
        model.fs.state[0].pressure] == 1e-5

    assert model.fs.state[0].scaling_factor[
        model.fs.state[0].flow_mass_comp["H2O"]] == 1
    assert model.fs.state[0].scaling_factor[
        model.fs.state[0].flow_mass_comp["A"]] == 1e5
    assert model.fs.state[0].scaling_factor[
        model.fs.state[0].flow_mass_comp["B"]] == 5e1
    assert model.fs.state[0].scaling_factor[
        model.fs.state[0].flow_mass_comp["C"]] == 1e5

    assert model.fs.state[0].scaling_factor[
        model.fs.state[0]._enth_dens_term] == 1e-5
    assert model.fs.state[0].scaling_factor[
        model.fs.state[0]._enth_flow_term] == 1e-4


@pytest.mark.component
def test_unit_consistency(model):
    assert_units_consistent(model)

    for e in model.fs.state[0].flow_mass_comp.values():
        assert_units_equivalent(e, pyunits.kg/pyunits.s)
    for e in model.fs.state[0]._enth_dens_term.values():
        assert_units_equivalent(e, pyunits.J/pyunits.m**3)
    for e in model.fs.state[0]._enth_flow_term.values():
        assert_units_equivalent(e, pyunits.J/pyunits.s)


@pytest.mark.component
def test_initialize_state_block(model):
    orig_fixed_vars = fixed_variables_set(model)
    orig_act_consts = activated_constraints_set(model)

    flags = model.fs.state.initialize(hold_state=True)

    assert degrees_of_freedom(model) == 0
    inter_fixed_vars = fixed_variables_set(model)
    for v in inter_fixed_vars:
        assert v.name in ['fs.state[0].flow_vol',
                          'fs.state[0].temperature',
                          'fs.state[0].pressure',
                          'fs.state[0].conc_mass_comp[A]',
                          'fs.state[0].conc_mass_comp[B]',
                          'fs.state[0].conc_mass_comp[C]']

    model.fs.state.release_state(flags)

    fin_fixed_vars = fixed_variables_set(model)
    fin_act_consts = activated_constraints_set(model)

    assert len(fin_act_consts) == len(orig_act_consts)
    assert len(fin_fixed_vars) == len(orig_fixed_vars)

    for c in fin_act_consts:
        assert c in orig_act_consts
    for v in fin_fixed_vars:
        assert v in orig_fixed_vars


@pytest.mark.component
def test_CV_integration(model):
    model.fs.cv = ControlVolume0DBlock(default={
            "property_package": model.fs.water_props})

    model.fs.cv.add_geometry()

    model.fs.cv.add_state_blocks(has_phase_equilibrium=True)

    model.fs.cv.add_material_balances(has_phase_equilibrium=True)

    model.fs.cv.add_energy_balances()

    model.fs.cv.add_momentum_balances()



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

"""Pretreatment flowsheet components"""

from pyomo.environ import ConcreteModel, TransformationFactory, Constraint
from pyomo.network import Arc
from idaes.core import FlowsheetBlock
from idaes.generic_models.unit_models import Feed, Translator
from idaes.core.util.scaling import (calculate_scaling_factors,
                                     constraint_scaling_transform,
                                     get_scaling_factor,
                                     constraint_autoscale_large_jac)
from idaes.core.util.initialization import propagate_state
from watertap.examples.flowsheets.full_treatment_train.flowsheet_components.chemistry import \
    pretreatment_stoich_softening_block as pssb
from watertap.examples.flowsheets.full_treatment_train.util import solve_with_user_scaling, check_dof


def build(m):
    """
    Builds softening pretreatment including specified feed and auxiliary equipment.
    """
    pretrt_port = {}

    pssb.build_stoich_softening_block(m)
    build_feed_block(m)

    # connect feed block
    m.fs.s_prtrt_feed_mixer = Arc(
        source=m.fs.feed.outlet, destination=m.fs.stoich_softening_mixer_unit.inlet_stream)

    # set model values
    pssb.set_stoich_softening_mixer_inlets(m, dosing_rate_of_lime_mg_per_s=500)
    pssb.fix_stoich_softening_mixer_lime_stream(m)

    # set ports
    pretrt_port['out'] = m.fs.stoich_softening_separator_unit.outlet_stream
    pretrt_port['waste'] = m.fs.stoich_softening_separator_unit.waste_stream

    return pretrt_port


def build_feed_block(m):
    m.fs.feed = Feed(default={'property_package': m.fs.stoich_softening_thermo_params})

    comp_list = ['Na', 'Ca', 'Mg', 'SO4', 'Cl']
    feed_flow_mass = 1  # kg/s
    feed_mass_frac_comp = {'Na': 11122e-6,
                           'Ca': 382e-6,
                           'Mg': 1394e-6,
                           'SO4': 2136e-6,
                           'Cl': 20316.88e-6}
    feed_mass_frac_comp['H2O'] = 1 - sum(x for x in feed_mass_frac_comp.values())

    mw_comp = {'H2O': 18.015e-3,
               'Na': 22.990e-3,
               'Ca': 40.078e-3,
               'Mg': 24.305e-3,
               'SO4': 96.06e-3,
               'Cl': 35.453e-3}

    feed_flow_mol_comp = {}
    for j in feed_mass_frac_comp:
        feed_flow_mol_comp[j] = feed_flow_mass * feed_mass_frac_comp[j] / mw_comp[j]

    feed_mole_frac_comp = {}
    for j in feed_flow_mol_comp:
        feed_mole_frac_comp[j] = feed_flow_mol_comp[j] / sum(feed_flow_mol_comp[jj] for jj in feed_flow_mol_comp)

    m.fs.feed.properties[0].mole_frac_comp['Ca(HCO3)2'].fix(feed_mole_frac_comp['Ca'])
    m.fs.feed.properties[0].mole_frac_comp['Ca(OH)2'].fix(0)
    m.fs.feed.properties[0].mole_frac_comp['CaCO3'].fix(0)
    m.fs.feed.properties[0].mole_frac_comp['H2O'].fix(feed_mole_frac_comp['H2O'])
    m.fs.feed.properties[0].mole_frac_comp['Mg(HCO3)2'].fix(feed_mole_frac_comp['Mg'])
    m.fs.feed.properties[0].mole_frac_comp['Mg(OH)2'].fix(0)
    m.fs.feed.properties[0].mole_frac_comp['NaCl'].fix(feed_mole_frac_comp['Na'])
    m.fs.feed.properties[0].mole_frac_comp['SO4_2-'].fix(feed_mole_frac_comp['SO4'])
    m.fs.feed.properties[0].mole_frac_comp['Cl_-'].fix(feed_mole_frac_comp['Cl'] - feed_mole_frac_comp['Na'])

    m.fs.feed.properties[0].flow_mol.fix(55.56)
    m.fs.feed.properties[0].pressure.fix(101325)
    m.fs.feed.properties[0].temperature.fix(298)


def build_tb(m):
    # build translator block
    m.fs.tb_pretrt_to_desal = Translator(
        default={"inlet_property_package": m.fs.stoich_softening_thermo_params,
                 "outlet_property_package": m.fs.prop_TDS})
    blk = m.fs.tb_pretrt_to_desal

    # add translator block constraints
    blk.eq_equal_temperature = Constraint(
        expr=blk.properties_in[0].temperature
             == blk.properties_out[0].temperature)
    blk.eq_equal_pressure = Constraint(
        expr=blk.properties_in[0].pressure
             == blk.properties_out[0].pressure)
    blk.eq_H2O_balance = Constraint(
        expr=blk.properties_in[0].flow_mol * blk.properties_in[0].mole_frac_comp['H2O']
             == blk.properties_out[0].flow_mol_phase_comp['Liq', 'H2O'])

    mw_comp = {'H2O': 18.015e-3,
               'Na': 22.990e-3,
               'Ca': 40.078e-3,
               'Mg': 24.305e-3,
               'SO4': 96.06e-3,
               'Cl': 35.453e-3}

    # TODO: this does not catch the Cl that is not associated with with Na
    blk.eq_TDS_balance = Constraint(
        expr=
        blk.properties_out[0].flow_mass_phase_comp['Liq', 'TDS']
        == blk.properties_in[0].flow_mol * blk.properties_in[0].mole_frac_comp['Ca(HCO3)2'] * mw_comp['Ca']
        + blk.properties_in[0].flow_mol * blk.properties_in[0].mole_frac_comp['Mg(HCO3)2'] * mw_comp['Mg']
        + blk.properties_in[0].flow_mol * blk.properties_in[0].mole_frac_comp['NaCl'] * (mw_comp['Na'] + mw_comp['Cl'])
        + blk.properties_in[0].flow_mol * blk.properties_in[0].mole_frac_comp['Cl_-'] * mw_comp['Cl']
        + blk.properties_in[0].flow_mol * blk.properties_in[0].mole_frac_comp['SO4_2-'] * mw_comp['SO4']
        )

    # scale translator block to get scaling factors
    calculate_scaling_factors(blk)
    constraint_scaling_transform(blk.eq_equal_temperature, get_scaling_factor(blk.properties_out[0].temperature))
    constraint_scaling_transform(blk.eq_equal_pressure, get_scaling_factor(blk.properties_out[0].pressure))
    constraint_scaling_transform(blk.eq_H2O_balance,
        get_scaling_factor(blk.properties_out[0].flow_mol_phase_comp['Liq', 'H2O']))
    constraint_scaling_transform( blk.eq_TDS_balance,
        get_scaling_factor(blk.properties_out[0].flow_mass_phase_comp['Liq', 'TDS']))

    # touch variables
    blk.properties_out[0].mass_frac_phase_comp


def scale(m):
    # mixer
    calculate_scaling_factors(m)
    constraint_autoscale_large_jac(m.fs.stoich_softening_mixer_unit)
    constraint_autoscale_large_jac(m.fs.stoich_softening_reactor_unit)
    constraint_autoscale_large_jac(m.fs.stoich_softening_separator_unit)

    # m.fs.feed.properties[0].scaling_factor.display()
    # assert False


def initialize(m):
    # initialize feed
    m.fs.feed.initialize()
    propagate_state(m.fs.s_prtrt_feed_mixer)
    # initializer mixer
    pssb.initialize_stoich_softening_mixer(m.fs.stoich_softening_mixer_unit, debug_out=False)
    # initializer reactor
    propagate_state(m.fs.stoich_softening_arc_mixer_to_reactor)
    pssb.initialize_stoich_softening_reactor(m.fs.stoich_softening_reactor_unit, debug_out=False)
    # initializer separator
    propagate_state(m.fs.stoich_softening_arc_reactor_to_separator)
    pssb.initialize_stoich_softening_separator(m.fs.stoich_softening_separator_unit, debug_out=False)


def display(m):
    m.fs.feed.report()
    pssb.display_results_of_stoich_softening_mixer(m.fs.stoich_softening_mixer_unit)
    pssb.display_results_of_stoich_softening_reactor(m.fs.stoich_softening_reactor_unit)
    pssb.display_results_of_stoich_softening_separator(m.fs.stoich_softening_separator_unit)


def solve():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    build(m)
    TransformationFactory("network.expand_arcs").apply_to(m)

    scale(m)

    initialize(m)

    check_dof(m)

    # solve
    solve_with_user_scaling(m)

    # display
    display(m)

    return m


if __name__ == "__main__":
    solve()



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

import pytest
from pyomo.environ import (ConcreteModel,
                           value,
                           Var,
                           Constraint,
                           assert_optimal_termination)
from pyomo.network import Port
from idaes.core import (FlowsheetBlock,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType)
from watertap.unit_models.nanofiltration_ZO import NanofiltrationZO
from idaes.generic_models.properties.core.generic.generic_property import GenericParameterBlock
from watertap.property_models.seawater_ion_generic import configuration
import watertap.examples.flowsheets.full_treatment_train.model_components.seawater_ion_prop_pack as props
from watertap.core.util.initialization import assert_no_degrees_of_freedom
from pyomo.util.check_units import assert_units_consistent

from idaes.core.util import get_solver
from idaes.core.util.model_statistics import (number_variables,
                                              number_total_constraints,
                                              number_unused_variables)
from idaes.core.util.testing import initialization_tester
from idaes.core.util.scaling import (calculate_scaling_factors,
                                     constraint_scaling_transform,
                                     unscaled_variables_generator,
                                     unscaled_constraints_generator,
                                     badly_scaled_var_generator)

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()

# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = props.PropParameterBlock()
    m.fs.unit = NanofiltrationZO(default={"property_package": m.fs.properties})

    assert len(m.fs.unit.config) == 8

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.material_balance_type == \
           MaterialBalanceType.useDefault
    assert m.fs.unit.config.energy_balance_type == \
           EnergyBalanceType.useDefault
    assert m.fs.unit.config.momentum_balance_type == \
           MomentumBalanceType.pressureTotal
    assert not m.fs.unit.config.has_pressure_change
    assert m.fs.unit.config.property_package is \
           m.fs.properties

@pytest.mark.unit
def test_option_has_pressure_change():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = props.PropParameterBlock()
    m.fs.unit = NanofiltrationZO(
        default={"property_package": m.fs.properties,
                 "has_pressure_change": True})

    assert isinstance(m.fs.unit.feed_side.deltaP, Var)
    assert isinstance(m.fs.unit.deltaP, Var)

class TestNanofiltration():
    @pytest.fixture(scope="class")
    def unit_frame(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = props.PropParameterBlock()

        m.fs.unit = NanofiltrationZO(default={"property_package": m.fs.properties})

        # fully specify system
        feed_flow_mass = 1
        feed_mass_frac = {'Na': 11122e-6,
                          'Ca': 382e-6,
                          'Mg': 1394e-6,
                          'SO4': 2136e-6,
                          'Cl': 20316.88e-6}
        m.fs.unit.feed_side.properties_in[0].flow_mass_phase_comp['Liq', 'H2O'].fix(
            feed_flow_mass * (1 - sum(x for x in feed_mass_frac.values())))
        for j in feed_mass_frac:
            m.fs.unit.feed_side.properties_in[0].flow_mass_phase_comp['Liq', j].fix(feed_flow_mass * feed_mass_frac[j])
        m.fs.unit.feed_side.properties_in[0].pressure.fix(101325)
        m.fs.unit.feed_side.properties_in[0].temperature.fix(298.15)

        m.fs.unit.flux_vol_solvent.fix(1.67e-6)
        m.fs.unit.area.fix(500)
        m.fs.unit.properties_permeate[0].pressure.fix(101325)

        m.fs.unit.rejection_phase_comp[0, 'Liq', 'Na'].fix(0.01)
        m.fs.unit.rejection_phase_comp[0, 'Liq', 'Ca'].fix(0.79)
        m.fs.unit.rejection_phase_comp[0, 'Liq', 'Mg'].fix(0.94)
        m.fs.unit.rejection_phase_comp[0, 'Liq', 'SO4'].fix(0.87)
        m.fs.unit.rejection_phase_comp[0, 'Liq', 'Cl'] = 0.15  # guess, but electroneutrality enforced below
        charge_comp = {'Na': 1, 'Ca': 2, 'Mg': 2, 'SO4': -2, 'Cl': -1}
        m.fs.unit.eq_electroneutrality = Constraint(
            expr=0 ==
                 sum(charge_comp[j]
                     * m.fs.unit.feed_side.properties_out[0].conc_mol_phase_comp['Liq', j]
                     for j in charge_comp))
        constraint_scaling_transform(m.fs.unit.eq_electroneutrality, 1)

        return m

    @pytest.mark.unit
    def test_build(self, unit_frame):
        m = unit_frame

        # test ports
        port_lst = ['inlet', 'retentate', 'permeate']
        for port_str in port_lst:
            port = getattr(m.fs.unit, port_str)
            assert len(port.vars) == 3  # number of state variables for NaCl property package
            assert isinstance(port, Port)

        # test statistics
        assert number_variables(m) == 89
        assert number_total_constraints(m) == 74
        assert number_unused_variables(m) == 1

    @pytest.mark.unit
    def test_dof(self, unit_frame):
        m = unit_frame
        assert_no_degrees_of_freedom(m)

    @pytest.mark.unit
    def test_calculate_scaling(self, unit_frame):
        m = unit_frame

        m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1, index=('Liq', 'H2O'))
        m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e2, index=('Liq', 'Na'))
        m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e4, index=('Liq', 'Ca'))
        m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e3, index=('Liq', 'Mg'))
        m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e3, index=('Liq', 'SO4'))
        m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e2, index=('Liq', 'Cl'))
        calculate_scaling_factors(m)

        # check that all variables have scaling factors
        unscaled_var_list = list(unscaled_variables_generator(m))
        assert len(unscaled_var_list) == 0
        # check that all constraints have been scaled
        unscaled_constraint_list = list(unscaled_constraints_generator(m))
        assert len(unscaled_constraint_list) == 0

    @pytest.mark.component
    def test_initialize(self, unit_frame):
        initialization_tester(unit_frame)

    @pytest.mark.component
    def test_var_scaling(self, unit_frame):
        m = unit_frame
        badly_scaled_var_lst = list(badly_scaled_var_generator(m))
        assert badly_scaled_var_lst == []

    @pytest.mark.component
    def test_solve(self, unit_frame):
        m = unit_frame
        solver.options = {'nlp_scaling_method': 'user-scaling'}
        results = solver.solve(m)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.component
    def test_conservation(self, unit_frame):
        m = unit_frame
        b = m.fs.unit
        comp_lst = m.fs.properties.component_list

        flow_mass_inlet = sum(
            b.feed_side.properties_in[0].flow_mass_phase_comp['Liq', j] for j in comp_lst)
        flow_mass_retentate = sum(
            b.feed_side.properties_out[0].flow_mass_phase_comp['Liq', j] for j in comp_lst)
        flow_mass_permeate = sum(
            b.properties_permeate[0].flow_mass_phase_comp['Liq', j] for j in comp_lst)

        assert (abs(value(flow_mass_inlet - flow_mass_retentate - flow_mass_permeate
                          )) <= 1e-5)

    @pytest.mark.component
    def test_solution(self, unit_frame):
        m = unit_frame
        assert (pytest.approx(0.8350, rel=1e-3) ==
                value(m.fs.unit.properties_permeate[0].flow_mass_phase_comp['Liq', 'H2O']))
        assert (pytest.approx(2.002, rel=1e-3) ==
                value(m.fs.unit.properties_permeate[0].conc_mol_phase_comp['Liq', 'Ca']))
        assert (pytest.approx(484.0, rel=1e-3) ==
                value(m.fs.unit.properties_permeate[0].conc_mol_phase_comp['Liq', 'Cl']))
        assert (pytest.approx(3.441, rel=1e-3) ==
                value(m.fs.unit.properties_permeate[0].conc_mol_phase_comp['Liq', 'Mg']))
        assert (pytest.approx(478.9, rel=1e-3) ==
                value(m.fs.unit.properties_permeate[0].conc_mol_phase_comp['Liq', 'Na']))
        assert (pytest.approx(2.891, rel=1e-3) ==
                value(m.fs.unit.properties_permeate[0].conc_mol_phase_comp['Liq', 'SO4']))
        assert (pytest.approx(0.86, rel=1e-3) ==
                value(m.fs.unit.recovery_vol_phase[0, 'Liq']))

    @pytest.mark.unit
    def test_report(self, unit_frame):
        unit_frame.fs.unit.report()

    @pytest.mark.component
    def test_NF_with_generic_property_model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.properties = GenericParameterBlock(default=configuration)
        m.fs.unit = NanofiltrationZO(default={
            "property_package": m.fs.properties,
            "has_pressure_change": False})

        # fully specify system
        m.fs.unit.inlet.flow_mol_phase_comp[0, 'Liq', 'Na_+'].fix(0.008845)
        m.fs.unit.inlet.flow_mol_phase_comp[0, 'Liq', 'Ca_2+'].fix(0.000174)
        m.fs.unit.inlet.flow_mol_phase_comp[0, 'Liq', 'Mg_2+'].fix(0.001049)
        m.fs.unit.inlet.flow_mol_phase_comp[0, 'Liq', 'SO4_2-'].fix(0.000407)
        m.fs.unit.inlet.flow_mol_phase_comp[0, 'Liq', 'Cl_-'].fix(0.010479)
        m.fs.unit.inlet.flow_mol_phase_comp[0, 'Liq', 'H2O'].fix(0.979046)
        m.fs.unit.feed_side.properties_in[0].pressure.fix(4e5)
        m.fs.unit.feed_side.properties_in[0].temperature.fix(298.15)

        m.fs.unit.flux_vol_solvent.fix(1.67e-6)
        m.fs.unit.recovery_vol_phase.fix(0.86)
        m.fs.unit.properties_permeate[0].pressure.fix(101325)

        m.fs.unit.rejection_phase_comp[0, 'Liq', 'Na_+'].fix(0.01)
        m.fs.unit.rejection_phase_comp[0, 'Liq', 'Ca_2+'].fix(0.79)
        m.fs.unit.rejection_phase_comp[0, 'Liq', 'Mg_2+'].fix(0.94)
        m.fs.unit.rejection_phase_comp[0, 'Liq', 'SO4_2-'].fix(0.87)
        m.fs.unit.rejection_phase_comp[0, 'Liq', 'Cl_-'] = 0.15  # guess, but electroneutrality enforced below
        charge_comp = {'Na_+': 1, 'Ca_2+': 2, 'Mg_2+': 2, 'SO4_2-': -2, 'Cl_-': -1}
        m.fs.unit.eq_electroneutrality = Constraint(
            expr=0 ==
                 sum(charge_comp[j]
                     * m.fs.unit.feed_side.properties_out[0].conc_mol_phase_comp['Liq', j]
                     for j in charge_comp))
        constraint_scaling_transform(m.fs.unit.eq_electroneutrality, 1)

        assert_units_consistent(m)

        assert_no_degrees_of_freedom(m)

        calculate_scaling_factors(m)

        initialization_tester(m)

        results = solver.solve(m)

        # Check for optimal solution
        assert_optimal_termination(results)

        assert (pytest.approx(0.868, rel=1e-3) ==
                value(m.fs.unit.properties_permeate[0].flow_mass_phase_comp['Liq', 'H2O'] /
                      m.fs.unit.feed_side.properties_in[0].flow_mass_phase_comp['Liq', 'H2O']))
        assert (pytest.approx(1.978, rel=1e-3) ==
                value(m.fs.unit.properties_permeate[0].conc_mol_phase_comp['Liq', 'Ca_2+']))
        assert (pytest.approx(479.1, rel=1e-3) ==
                value(m.fs.unit.properties_permeate[0].conc_mol_phase_comp['Liq', 'Cl_-']))
        assert (pytest.approx(3.407, rel=1e-3) ==
                value(m.fs.unit.properties_permeate[0].conc_mol_phase_comp['Liq', 'Mg_2+']))
        assert (pytest.approx(473.9, rel=1e-3) ==
                value(m.fs.unit.properties_permeate[0].conc_mol_phase_comp['Liq', 'Na_+']))
        assert (pytest.approx(2.864, rel=1e-3) ==
                value(m.fs.unit.properties_permeate[0].conc_mol_phase_comp['Liq', 'SO4_2-']))




###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

from pyomo.environ import (
    Block,
    ConcreteModel,
    Constraint,
    Expression,
    Var,
    Param,
    value,
    TransformationFactory,
    units as pyunits,
)
import watertap.examples.flowsheets.full_treatment_train.flowsheet_components.financials as financials
from watertap.examples.flowsheets.full_treatment_train.flowsheet_components import (
    feed_block,
)
from watertap.examples.flowsheets.full_treatment_train.model_components import (
    unit_separator,
    unit_0DRO,
    unit_1DRO,
    property_models,
)

from watertap.examples.flowsheets.full_treatment_train.flowsheet_components.desalination import (
    build_desalination,
    solve_desalination,
    scale_desalination,
    initialize_desalination,
    display_desalination,
)

# from watertap.examples.flowsheets.full_treatment_train.analysis.flowsheet_softening_two_stage import build, scale, initialize
from watertap.examples.flowsheets.full_treatment_train.util import (
    solve_with_user_scaling,
    check_dof,
)
import idaes.core.util.scaling as iscale
from idaes.core import FlowsheetBlock


cost_capacity_flag = False


def build_costing(m, module=financials, **kwargs):
    """Add costing to a given flowsheet. This function will:
        1) call the get_costing method for each unit model (note: unit model must have a get_costing method
           to be detected), and
        2) call get_system_costing which will tally up all capex and opex for each process

    Args:
        m: model
        module: financials module
    """
    crf = m.fs.costing_param.factor_capital_annualization

    # call get_costing for each unit model
    # TODO: add in other components as they become available
    m.fs.costing = Block()

    # Nanofiltration
    if hasattr(m.fs, "NF"):
        if kwargs["NF_type"] == "ZO":
            m.fs.NF.get_costing(module=module, section="pretreatment")
        elif kwargs["NF_type"] == "Sep":
            raise NotImplementedError(
                "get_costing will not be implemented for the NF separator model."
            )
    if hasattr(m.fs, "pump_NF"):
        m.fs.pump_NF.get_costing(
            module=module,
            section="pretreatment",
            pump_type="low pressure",
            cost_capacity=cost_capacity_flag,
        )

    # Reverse Osmosis
    if hasattr(m.fs, "RO"):
        if kwargs["RO_type"] == "0D" or kwargs["RO_type"] == "1D":
            m.fs.RO.get_costing(module=module, section="primary")
        elif kwargs["RO_type"] == "Sep":
            raise NotImplementedError(
                "get_costing will not be implemented for the RO separator model."
            )

    # Stage 2 RO
    if hasattr(m.fs, "RO2"):
        m.fs.RO2.get_costing(module=module, RO_type="high pressure", section="primary")

    # Pump
    if hasattr(m.fs, "pump_RO"):
        m.fs.pump_RO.get_costing(
            module=module, section="primary", pump_type="high pressure"
        )

    # Stage 2 pump
    if hasattr(m.fs, "pump_RO2"):
        m.fs.pump_RO2.get_costing(
            module=module, section="primary", pump_type="high pressure"
        )

    # ERD
    if hasattr(m.fs, "ERD"):
        m.fs.ERD.get_costing(
            module=module, section="primary", pump_type="pressure exchanger"
        )

    # Pretreatment
    if hasattr(m.fs, "stoich_softening_mixer_unit"):
        m.fs.stoich_softening_mixer_unit.get_costing(
            module=module,
            section="pretreatment",
            mixer_type="lime_softening",
            cost_capacity=cost_capacity_flag,
        )
        m.fs.lime_softening_unit_capex = Expression(
            expr=m.fs.stoich_softening_mixer_unit.costing.capital_cost
            / m.fs.annual_water_production
            * crf
        )
        m.fs.lime_softening_unit_opex = Expression(
            expr=m.fs.stoich_softening_mixer_unit.costing.operating_cost
            / m.fs.annual_water_production
        )
    else:
        m.fs.lime_softening_unit_capex = Expression(expr=0)
        m.fs.lime_softening_unit_opex = Expression(expr=0)

    # Post-treatment
    if hasattr(m.fs, "ideal_naocl_mixer_unit"):
        # print('FOUND CHLORINATION UNIT')
        m.fs.ideal_naocl_mixer_unit.get_costing(
            module=module,
            section="post_treatment",
            mixer_type="naocl_mixer",
            cost_capacity=cost_capacity_flag,
        )
        m.fs.chlorination_unit_capex = Expression(
            expr=m.fs.ideal_naocl_mixer_unit.costing.capital_cost
            / m.fs.annual_water_production
            * crf
        )
        m.fs.chlorination_unit_opex = Expression(
            expr=m.fs.ideal_naocl_mixer_unit.costing.operating_cost
            / m.fs.annual_water_production
        )
    else:
        m.fs.chlorination_unit_capex = Expression(expr=0)
        m.fs.chlorination_unit_opex = Expression(expr=0)

    if hasattr(m.fs, "mixer_permeate"):
        m.fs.mixer_permeate.get_costing(
            module=module, section="primary", cost_capacity=cost_capacity_flag
        )

    # call get_system_costing for whole flowsheet
    module.get_system_costing(m.fs)

    # apply scaling to cost variables and constraints
    scale_costing(m)


def scale_costing(self):
    for b_unit in self.component_objects(Block, descend_into=True):
        if hasattr(b_unit, "costing"):
            base = b_unit.costing
            for var in base.component_objects(Var):
                if iscale.get_scaling_factor(var) is None:
                    iscale.set_scaling_factor(var, 1e-3)
                for con in base.component_objects(Constraint):
                    sf = iscale.get_scaling_factor(var)
                    iscale.constraint_scaling_transform(con, sf)


def display_costing(m):
    crf = m.fs.costing_param.factor_capital_annualization
    if not hasattr(m.fs, "pump_RO2"):
        m.fs.pump_RO2 = Block()
        m.fs.pump_RO2.costing = Block()
        m.fs.pump_RO2.costing.operating_cost = Param(initialize=0)
    if not hasattr(m.fs, "NF"):
        m.fs.NF = Block()
        m.fs.NF.costing = Block()
        m.fs.NF.costing.operating_cost = Param(initialize=0)
    if not hasattr(m.fs, "RO2"):
        m.fs.RO2 = Block()
        m.fs.RO2.costing = Block()
        m.fs.RO2.costing.operating_cost = Param(initialize=0)

    # UNITS FOR ALL COST COMPONENTS [=] $/m3 of permeate water produced
    cost_dict = {
        "LCOW": m.fs.costing.LCOW,  # Total LCOW
        "Total CAPEX": m.fs.costing.investment_cost_total
        * crf
        / m.fs.annual_water_production,  # Direct + Indirect CAPEX
        "Direct CAPEX": m.fs.costing.capital_cost_total
        * crf
        / m.fs.annual_water_production,  # Direct CAPEX for all system components
        "Indirect CAPEX": (
            m.fs.costing.investment_cost_total - m.fs.costing.capital_cost_total
        )
        * crf
        / m.fs.annual_water_production,  # Indirect CAPEX for miscellaneous items
        "Total OPEX": m.fs.costing.operating_cost_total
        / m.fs.annual_water_production,  # Total OPEX
        "Labor & Maintenance Costs": m.fs.costing.operating_cost_labor_maintenance
        / m.fs.annual_water_production,
        "Total Electricity Cost": m.fs.costing.electricity_cost_total
        / m.fs.annual_water_production,
        "Total Membrane Replacement Cost": (
            m.fs.NF.costing.operating_cost
            + m.fs.RO.costing.operating_cost
            + m.fs.RO2.costing.operating_cost
        )
        / m.fs.annual_water_production,
        "Total Pretreatment Cost": m.fs.costing.pretreatment_cost_total
        / m.fs.annual_water_production,
        "Total Primary Cost": m.fs.costing.primary_cost_total
        / m.fs.annual_water_production,
        "Total Post-treatment Cost": m.fs.costing.post_treatment_cost_total
        / m.fs.annual_water_production,
        "Lime softener CAPEX": m.fs.lime_softening_unit_capex,
        "Lime softener OPEX": m.fs.lime_softening_unit_opex,
        "Chlorination CAPEX": m.fs.chlorination_unit_capex,
        "Chlorination OPEX": m.fs.chlorination_unit_opex,
    }

    for item, val in cost_dict.items():
        print(f"{item} = {value(val)}")

    return cost_dict



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

from pyomo.environ import ConcreteModel, Objective, Expression, Constraint, TransformationFactory, value, Param, Block
from pyomo.environ import units as pyunits
from pyomo.network import Arc
from pyomo.util import infeasible
from idaes.core import FlowsheetBlock
from idaes.core.util.scaling import (calculate_scaling_factors,
                                     constraint_autoscale_large_jac)
from idaes.core.util.initialization import propagate_state
from watertap.examples.flowsheets.full_treatment_train.flowsheet_components import (pretreatment_NF,
                                                                             desalination,
                                                                             translator_block,
                                                                             feed_block,
                                                                             gypsum_saturation_index,
                                                                             costing,
                                                                             financials)
from watertap.examples.flowsheets.full_treatment_train.model_components import property_models
from watertap.examples.flowsheets.full_treatment_train.util import (solve_with_user_scaling,
                                                             check_dof,
                                                             check_build,
                                                             check_scaling)

"""Flowsheet example that satisfy minimum viable product requirements"""
def build_flowsheet_mvp_NF(m, **kwargs):
    """
    Build a flowsheet with NF pretreatment and RO.
    """
    # set up keyword arguments for the sections of treatment train
    kwargs_pretreatment = {k: kwargs[k] for k in ('has_bypass', 'NF_type', 'NF_base',)}
    kwargs_desalination = {k: kwargs[k] for k in ('has_desal_feed', 'is_twostage', 'has_ERD',
                                                  'RO_type', 'RO_base', 'RO_level',)}
    # build flowsheet
    property_models.build_prop(m, base='ion')
    pretrt_port = pretreatment_NF.build_pretreatment_NF(m, **kwargs_pretreatment)

    property_models.build_prop(m, base=kwargs['RO_base'])
    desal_port = desalination.build_desalination(m, **kwargs_desalination)

    property_models.build_prop(m, base='eNRTL')

    translator_block.build_tb(m, base_inlet=kwargs['NF_base'],
                              base_outlet=kwargs['RO_base'],
                              name_str='tb_pretrt_to_desal')

    # set up Arcs between pretreatment and desalination
    m.fs.s_pretrt_tb = Arc(source=pretrt_port['out'], destination=m.fs.tb_pretrt_to_desal.inlet)
    m.fs.s_tb_desal = Arc(source=m.fs.tb_pretrt_to_desal.outlet, destination=desal_port['in'])

    # add gypsum saturation index calculations
    gypsum_saturation_index.build(m, section='desalination', **kwargs_desalination)
    gypsum_saturation_index.build(m, section='pretreatment', **kwargs_desalination)

    # new initialization
    if kwargs['NF_type'] == 'ZO':
        m.fs.NF.area.fix(175)
    if kwargs['has_bypass']:
        m.fs.splitter.split_fraction[0, 'bypass'].fix(0.50)
    m.fs.RO.area.fix(80)
    m.fs.pump_RO.control_volume.properties_out[0].pressure.fix(60e5)
    if kwargs['is_twostage']:
        m.fs.RO2.area.fix(20)
        m.fs.pump_RO2.control_volume.properties_out[0].pressure.fix(90e5)

    # touch some properties used in optimization
    if kwargs['is_twostage']:
        product_water_sb = m.fs.mixer_permeate.mixed_state[0]
        if kwargs['RO_type'] == '0D':
            RO_waste_sb = m.fs.RO2.feed_side.properties_out[0]
        elif kwargs['RO_type'] == '1D':
            RO_waste_sb = m.fs.RO2.feed_side.properties[0, 1]
    else:
        if kwargs['RO_type'] == '0D':
            product_water_sb = m.fs.RO.permeate_side.properties_mixed[0]
            RO_waste_sb = m.fs.RO.feed_side.properties_out[0]
        elif kwargs['RO_type'] == '1D':
            product_water_sb = m.fs.RO.mixed_permeate[0]
            RO_waste_sb = m.fs.RO.feed_side.properties[0, 1]

    # NOTE: Building the costing here means it gets
    #       initialized during the simulation phase.
    #       This helps model stability.
    m.fs.feed.properties[0].flow_vol
    m.fs.feed.properties[0].conc_mol_phase_comp['Liq', 'Ca']

    m.fs.tb_pretrt_to_desal.properties_in[0].flow_vol
    m.fs.tb_pretrt_to_desal.properties_in[0].conc_mol_phase_comp['Liq', 'Ca']

    product_water_sb.flow_vol
    RO_waste_sb.flow_vol

    m.fs.system_recovery = Expression(
        expr=product_water_sb.flow_vol / m.fs.feed.properties[0].flow_vol)
    m.fs.total_work = Expression(expr=m.fs.pump_RO.work_mechanical[0] +
                                    (m.fs.pump_RO2.work_mechanical[0] if kwargs['is_twostage'] else 0.))

    # need load factor from costing_param_block for annual_water_production
    financials.add_costing_param_block(m.fs)
    # annual water production
    m.fs.annual_water_production = Expression(
        expr=pyunits.convert(product_water_sb.flow_vol, to_units=pyunits.m ** 3 / pyunits.year)
             * m.fs.costing_param.load_factor)
    costing.build_costing(m, module=financials, **kwargs)

    return m

def set_up_optimization(m, system_recovery=0.7, **kwargs_flowsheet):
    is_twostage = kwargs_flowsheet['is_twostage']

    # scale
    calculate_scaling_factors(m)

    # unfix variables
    m.fs.splitter.split_fraction[0, 'bypass'].unfix()
    m.fs.splitter.split_fraction[0, 'bypass'].setlb(0.001)
    m.fs.splitter.split_fraction[0, 'bypass'].setub(0.99)

    m.fs.NF.area.unfix()
    m.fs.NF.area.setlb(10)
    m.fs.NF.area.setub(1000)

    m.fs.pump_RO.control_volume.properties_out[0].pressure.unfix()
    m.fs.pump_RO.control_volume.properties_out[0].pressure.setlb(20e5)
    m.fs.pump_RO.control_volume.properties_out[0].pressure.setub(75e5)

    m.fs.RO.area.unfix()
    m.fs.RO.area.setlb(10)
    m.fs.RO.area.setub(300)

    # Set lower bound for water flux at the RO outlet, based on a minimum net driving pressure, NDPmin
    m.fs.RO.NDPmin = Param(initialize=1e5, mutable=True, units=pyunits.Pa)

    if kwargs_flowsheet['RO_type'] == '0D':
        m.fs.RO.flux_mass_io_phase_comp[0, 'out', 'Liq', 'H2O'].setlb(value(m.fs.RO.A_comp[0, 'H2O']
                                                                            * m.fs.RO.dens_solvent
                                                                            * m.fs.RO.NDPmin))
    elif kwargs_flowsheet['RO_type'] == '1D':
        m.fs.RO.flux_mass_phase_comp[0, 1, 'Liq', 'H2O'].setlb(value(m.fs.RO.A_comp[0, 'H2O']
                                                                     * m.fs.RO.dens_solvent
                                                                     * m.fs.RO.NDPmin))

    if is_twostage:
        m.fs.max_allowable_pressure = Param(initialize=120e5, mutable=True, units=pyunits.pascal)
        m.fs.pump_RO2.control_volume.properties_out[0].pressure.unfix()
        m.fs.pump_RO2.control_volume.properties_out[0].pressure.setlb(20e5)
        m.fs.pump_RO2.control_volume.properties_out[0].pressure.setub(m.fs.max_allowable_pressure)

        m.fs.RO2.area.unfix()
        m.fs.RO2.area.setlb(10)
        m.fs.RO2.area.setub(300)

        # Set lower bound for water flux at the RO outlet, based on a minimum net driving pressure, NDPmin
        m.fs.RO2.NDPmin = Param(initialize=1e5, mutable=True, units=pyunits.Pa)
        if kwargs_flowsheet['RO_type'] == '0D':
            m.fs.RO2.flux_mass_io_phase_comp[0, 'out', 'Liq', 'H2O'].setlb(value(m.fs.RO2.A_comp[0, 'H2O']
                                                                                 * m.fs.RO2.dens_solvent
                                                                                 * m.fs.RO2.NDPmin))
        elif kwargs_flowsheet['RO_type'] == '1D':
            m.fs.RO2.flux_mass_phase_comp[0, 1, 'Liq', 'H2O'].setlb(value(m.fs.RO2.A_comp[0, 'H2O']
                                                                          * m.fs.RO2.dens_solvent
                                                                          * m.fs.RO2.NDPmin))

    # add additional constraints
    # fixed system recovery
    m.fs.system_recovery_target = Param(initialize=system_recovery, mutable=True)
    m.fs.system_recovery_tol = Param(initialize=5e-3, mutable=True)
    m.fs.eq_system_recovery = Constraint(
        expr=(m.fs.system_recovery_target,
                m.fs.system_recovery,
                m.fs.system_recovery_target+m.fs.system_recovery_tol))

    # saturation index
    m.fs.max_saturation_index = Param(initialize=1.0, mutable=True)
    m.fs.eq_max_saturation_index_desal = Constraint(
        expr=m.fs.desal_saturation.saturation_index <= m.fs.max_saturation_index)

    m.fs.max_conc_factor_target = Param(initialize=3.5, mutable=True)
    m.fs.eq_max_conc_NF = Constraint(
        expr=m.fs.NF.feed_side.properties_out[0].mass_frac_phase_comp['Liq', 'Ca']
        <= m.fs.max_conc_factor_target * m.fs.feed.properties[0].mass_frac_phase_comp['Liq', 'Ca'])

    # set objective
    m.fs.objective = Objective(expr=m.fs.costing.LCOW)

    # set additional constraints to limit local minima
    # NOTE: doesn't seem necessary with new objective
    if False:
        m.fs.inequality_RO_area = Constraint(expr=m.fs.RO.area >= m.fs.RO2.area)
        min_pressure_increase = 1e5
        m.fs.inequality_RO_pressure = Constraint(
            expr=m.fs.pump_RO.control_volume.properties_out[0].pressure + min_pressure_increase
                 <= m.fs.pump_RO2.control_volume.properties_out[0].pressure)
        m.fs.inequality_RO_permeate = Constraint(
            expr=m.fs.RO.permeate_side.properties_mixed[0].flow_vol_phase['Liq']
            >= m.fs.RO2.permeate_side.properties_mixed[0].flow_vol_phase['Liq'])

    check_dof(m, dof_expected=6 if is_twostage else 4)
    # solve_with_user_scaling(m, tee=False, fail_flag=True)


def optimize(m):
    solve_with_user_scaling(m, tee=False, fail_flag=True)


def solve_flowsheet_mvp_NF(**kwargs):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    build_flowsheet_mvp_NF(m, **kwargs)
    TransformationFactory("network.expand_arcs").apply_to(m)

    # scale
    pretreatment_NF.scale_pretreatment_NF(m, **kwargs)
    calculate_scaling_factors(m.fs.tb_pretrt_to_desal)
    desalination.scale_desalination(m, **kwargs)
    calculate_scaling_factors(m)

    # initialize
    optarg = {'nlp_scaling_method': 'user-scaling'}
    pretreatment_NF.initialize_pretreatment_NF(m, **kwargs)
    m.fs.pretrt_saturation.properties.initialize(optarg=optarg)
    propagate_state(m.fs.s_pretrt_tb)
    m.fs.tb_pretrt_to_desal.initialize(optarg=optarg)
    propagate_state(m.fs.s_tb_desal)
    desalination.initialize_desalination(m, **kwargs)
    m.fs.desal_saturation.properties.initialize()

    # check_build(m)
    # check_scaling(m)

    check_dof(m)
    solve_with_user_scaling(m, tee=False, fail_flag=True)

    pretreatment_NF.display_pretreatment_NF(m, **kwargs)
    m.fs.tb_pretrt_to_desal.report()
    desalination.display_desalination(m, **kwargs)
    print('desalination solubility index:', value(m.fs.desal_saturation.saturation_index))
    print('pretreatment solubility index:', value(m.fs.pretrt_saturation.saturation_index))
    print('water recovery:', value(m.fs.system_recovery))
    print('LCOW:', value(m.fs.costing.LCOW))
    print('CP modulus:', value(m.fs.desal_saturation.cp_modulus))

    return m

def solve_optimization(system_recovery=0.75, **kwargs_flowsheet):

    m = solve_flowsheet_mvp_NF(**kwargs_flowsheet)

    print('\n****** Optimization *****\n')
    set_up_optimization(m, system_recovery=system_recovery, **kwargs_flowsheet)
    optimize(m)

    pretreatment_NF.display_pretreatment_NF(m, **kwargs_flowsheet)
    m.fs.tb_pretrt_to_desal.report()
    desalination.display_desalination(m, **kwargs_flowsheet)
    costing.display_costing(m)
    print('desalination saturation index:', value(m.fs.desal_saturation.saturation_index))
    print('pretreatment saturation index:', value(m.fs.pretrt_saturation.saturation_index))
    print('pretreatment Ca concentration factor:', value(
        m.fs.NF.feed_side.properties_out[0].mass_frac_phase_comp['Liq', 'Ca'] /
        m.fs.feed.properties[0].mass_frac_phase_comp['Liq', 'Ca']))
    print('water recovery:', value(m.fs.system_recovery))
    print('CP modulus:', value(m.fs.desal_saturation.cp_modulus))
    return m


if __name__ == "__main__":
    import sys
    kwargs_flowsheet = {
        'has_bypass': True, 'has_desal_feed': False, 'is_twostage': True, 'has_ERD': True,
        'NF_type': 'ZO', 'NF_base': 'ion',
        'RO_type': '0D', 'RO_base': 'TDS', 'RO_level': 'detailed'}
    if len(sys.argv) == 1:
        m = solve_flowsheet_mvp_NF(**kwargs_flowsheet)
    else:
        m = solve_optimization(system_recovery=float(sys.argv[1]), **kwargs_flowsheet)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
import pytest
from watertap.examples.edb.the_basics import (
    run_the_basics_with_mockdb,
    run_the_basics_alt_with_mockdb,
    run_the_basics_dummy_rxn_with_mockdb,
)
from watertap.examples.edb.simple_acid import run_simple_acid_with_mockdb
from watertap.examples.edb.vapor_liquid_equilibrium import run_vap_liq_with_mockdb
from watertap.examples.edb.solid_precipitation_reactions import (
    run_sol_liq_with_mockdb,
    run_liq_only_with_mockdb,
)

# Import pyomo methods to check the system units
from pyomo.util.check_units import assert_units_consistent
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
)
from idaes.core import AqueousPhase, VaporPhase, SolidPhase
from idaes.core.components import Solvent, Solute, Cation, Anion, Component

__author__ = "Austin Ladshaw"

@pytest.mark.component
def test_the_basics(edb):
    assert run_the_basics_with_mockdb(edb) == True

@pytest.mark.component
def test_the_basics_inherent(edb):
    assert run_the_basics_alt_with_mockdb(edb) == True

@pytest.mark.component
def test_the_basics_inherent_with_dummy(edb):
    assert run_the_basics_dummy_rxn_with_mockdb(edb) == True

@pytest.mark.component
def test_simple_acid(edb):
    model = run_simple_acid_with_mockdb(edb)

    # Check model stats
    assert_units_consistent(model)
    assert degrees_of_freedom(model) == 8
    assert len(model.fs.unit.inlet) == 1
    assert len(model.fs.unit.inlet.flow_mol_phase_comp) == 6

    # Check for correct species and phases
    assert hasattr(model.fs.thermo_params, 'component_list')
    assert len(model.fs.thermo_params.component_list) == 6
    assert 'H2O' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.H2O, Solvent)
    assert 'H_+' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.component('H_+'), Cation)
    assert 'OH_-' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.component('OH_-'), Anion)
    assert 'HCO3_-' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.component('HCO3_-'), Anion)
    assert 'CO3_2-' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.component('CO3_2-'), Anion)
    assert 'H2CO3' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.H2CO3, Solute)

    assert hasattr(model.fs.thermo_params, 'phase_list')
    assert len(model.fs.thermo_params.phase_list) == 1
    assert isinstance(model.fs.thermo_params.Liq, AqueousPhase)

    # Check for correct reactions
    assert len(model.fs.rxn_params.reaction_idx) == 3
    assert 'H2O_Kw' in model.fs.rxn_params.reaction_idx
    assert 'H2CO3_Ka1' in model.fs.rxn_params.reaction_idx
    assert 'H2CO3_Ka2' in model.fs.rxn_params.reaction_idx

@pytest.mark.component
def test_vapor_liquid_equilibrium(edb):
    model = run_vap_liq_with_mockdb(edb)

    # Check model stats
    assert_units_consistent(model)
    assert degrees_of_freedom(model) == 11

    # Check for correct species and phases
    assert hasattr(model.fs.thermo_params, 'component_list')
    assert len(model.fs.thermo_params.component_list) == 7
    assert 'H2O' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.H2O, Solvent)
    assert 'CO2' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.CO2, Solute)
    assert 'H_+' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.component('H_+'), Cation)
    assert 'OH_-' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.component('OH_-'), Anion)
    assert 'HCO3_-' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.component('HCO3_-'), Anion)
    assert 'CO3_2-' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.component('CO3_2-'), Anion)
    assert 'H2CO3' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.H2CO3, Solute)

    assert hasattr(model.fs.thermo_params, 'phase_list')
    assert len(model.fs.thermo_params.phase_list) == 2
    assert isinstance(model.fs.thermo_params.Liq, AqueousPhase)
    assert isinstance(model.fs.thermo_params.Vap, VaporPhase)

    # Check for correct reactions
    assert len(model.fs.rxn_params.reaction_idx) == 4
    assert 'H2O_Kw' in model.fs.rxn_params.reaction_idx
    assert 'H2CO3_Ka1' in model.fs.rxn_params.reaction_idx
    assert 'H2CO3_Ka2' in model.fs.rxn_params.reaction_idx
    assert 'CO2_to_H2CO3' in model.fs.rxn_params.reaction_idx

@pytest.mark.component
def test_solid_precipitation_reactions(edb):
    model = run_sol_liq_with_mockdb(edb)

    # Check model stats
    assert_units_consistent(model)
    assert degrees_of_freedom(model) == 8

    # Check for correct species and phases
    assert hasattr(model.fs.thermo_params, 'component_list')
    assert len(model.fs.thermo_params.component_list) == 6
    assert 'H2O' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.H2O, Solvent)
    assert 'H_+' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.component('H_+'), Cation)
    assert 'OH_-' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.component('OH_-'), Anion)
    assert 'Ca_2+' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.component('Ca_2+'), Cation)
    assert 'CaOH_+' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.component('CaOH_+'), Cation)
    assert 'Ca[OH]2' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.component('Ca[OH]2'), Component)

    assert hasattr(model.fs.thermo_params, 'phase_list')
    assert len(model.fs.thermo_params.phase_list) == 2
    assert isinstance(model.fs.thermo_params.Liq, AqueousPhase)
    assert isinstance(model.fs.thermo_params.Sol, SolidPhase)

    # Check for correct reactions
    assert len(model.fs.rxn_params.reaction_idx) == 3
    assert 'H2O_Kw' in model.fs.rxn_params.reaction_idx
    assert 'CaOH_K' in model.fs.rxn_params.reaction_idx
    assert 'CaOH2_Ksp' in model.fs.rxn_params.reaction_idx

    # Check for correct reaction order on solid species 
    assert model.fs.rxn_params.reaction_CaOH2_Ksp.reaction_order[('Sol', 'Ca[OH]2')].value == 0

@pytest.mark.component
def test_solid_precipitation_reactions_liq_only(edb):
    model = run_liq_only_with_mockdb(edb)

    # Check model stats
    assert_units_consistent(model)
    assert degrees_of_freedom(model) == 8

    # Check for correct species and phases
    assert hasattr(model.fs.thermo_params, 'component_list')
    assert len(model.fs.thermo_params.component_list) == 6
    assert 'H2O' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.H2O, Solvent)
    assert 'H_+' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.component('H_+'), Cation)
    assert 'OH_-' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.component('OH_-'), Anion)
    assert 'Ca_2+' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.component('Ca_2+'), Cation)
    assert 'CaOH_+' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.component('CaOH_+'), Cation)
    assert 'Ca[OH]2' in model.fs.thermo_params.component_list
    assert isinstance(model.fs.thermo_params.component('Ca[OH]2'), Component)

    assert hasattr(model.fs.thermo_params, 'phase_list')
    assert len(model.fs.thermo_params.phase_list) == 2
    assert isinstance(model.fs.thermo_params.Liq, AqueousPhase)
    assert isinstance(model.fs.thermo_params.Sol, SolidPhase)

    # Check for correct reactions
    assert len(model.fs.rxn_params.reaction_idx) == 2
    assert 'H2O_Kw' in model.fs.rxn_params.reaction_idx
    assert 'CaOH_K' in model.fs.rxn_params.reaction_idx



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
import pytest
import watertap.property_models.seawater_prop_pack as props
from idaes.generic_models.properties.tests.test_harness import \
    PropertyTestHarness as PropertyTestHarness_idaes
from watertap.property_models.tests.property_test_harness import \
    (PropertyTestHarness, PropertyRegressionTest, PropertyCalculateStateTest)

# -----------------------------------------------------------------------------
class TestSeawaterProperty_idaes(PropertyTestHarness_idaes):
    def configure(self):
        self.prop_pack = props.SeawaterParameterBlock
        self.param_args = {}
        self.prop_args = {}
        self.has_density_terms = False


class TestSeawaterProperty(PropertyTestHarness):
    def configure(self):
        self.prop_pack = props.SeawaterParameterBlock
        self.param_args = {}
        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1,
                             ('flow_mass_phase_comp', ('Liq', 'TDS')): 1e2}
        self.stateblock_statistics = {'number_variables': 25,
                                      'number_total_constraints': 21,
                                      'number_unused_variables': 1,  # pressure is unused
                                      'default_degrees_of_freedom': 3}  # 4 state vars, but pressure is not active
        self.default_solution = {('mass_frac_phase_comp', ('Liq', 'H2O')): 0.965,
                                 ('mass_frac_phase_comp', ('Liq', 'TDS')): 0.035,
                                 ('dens_mass_phase', 'Liq'): 1023.5,
                                 ('dens_mass_solvent', None): 996.9,
                                 ('flow_vol_phase', 'Liq'): 9.770e-4,
                                 ('conc_mass_phase_comp', ('Liq', 'H2O')): 987.7,
                                 ('conc_mass_phase_comp', ('Liq', 'TDS')): 35.82,
                                 ('flow_mol_phase_comp', ('Liq', 'H2O')): 53.57,
                                 ('flow_mol_phase_comp', ('Liq', 'TDS')): 1.1145,
                                 ('mole_frac_phase_comp', ('Liq', 'H2O')): 0.9796,
                                 ('mole_frac_phase_comp', ('Liq', 'TDS')): 2.038e-2,
                                 ('molality_comp', 'TDS'): 1.155,
                                 ('visc_d_phase', 'Liq'): 9.588e-4,
                                 ('osm_coeff', None): 0.9068,
                                 ('pressure_osm', None): 2.588e6,
                                 ('enth_mass_phase', 'Liq'): 9.9765e4,
                                 ('pressure_sat', None): 3111,
                                 ('cp_phase', 'Liq'): 4001,
                                 ('therm_cond_phase', 'Liq'): 0.6086,
                                 ('dh_vap', None): 2.356e6,
                                 ('diffus_phase', 'Liq'): 1.471e-9}


@pytest.mark.component
class TestSeawaterPropertySolution_1(PropertyRegressionTest):
    def configure(self):
        self.prop_pack = props.SeawaterParameterBlock
        self.param_args = {}

        self.solver = 'ipopt'
        self.optarg = {'nlp_scaling_method': 'user-scaling'}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1,
                             ('flow_mass_phase_comp', ('Liq', 'TDS')): 1e2}
        self.state_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 0.99,
                           ('flow_mass_phase_comp', ('Liq', 'TDS')): 0.01,
                           ('temperature', None): 273.15 + 50,
                           ('pressure', None): 2e5}
        self.regression_solution = {('mass_frac_phase_comp', ('Liq', 'H2O')): 0.99,
                                    ('mass_frac_phase_comp', ('Liq', 'TDS')): 0.01,
                                    ('dens_mass_phase', 'Liq'): 995.4,
                                    ('dens_mass_solvent', None): 988.0,
                                    ('flow_vol_phase', 'Liq'): 1.005e-3,
                                    ('conc_mass_phase_comp', ('Liq', 'H2O')): 985.5,
                                    ('conc_mass_phase_comp', ('Liq', 'TDS')): 9.954,
                                    ('flow_mol_phase_comp', ('Liq', 'H2O')): 54.95,
                                    ('flow_mol_phase_comp', ('Liq', 'TDS')): 0.3184,
                                    ('mole_frac_phase_comp', ('Liq', 'H2O')): 0.9942,
                                    ('mole_frac_phase_comp', ('Liq', 'TDS')): 5.761e-3,
                                    ('molality_comp', 'TDS'): 0.3216,
                                    ('visc_d_phase', 'Liq'): 5.596e-4,
                                    ('osm_coeff', None): 0.9029,
                                    ('pressure_osm', None): 7.710e5,
                                    ('enth_mass_phase', 'Liq'): 2.066e5,
                                    ('pressure_sat', None): 1.229e4,
                                    ('cp_phase', 'Liq'): 4.130e3,
                                    ('therm_cond_phase', 'Liq'): 0.6400,
                                    ('dh_vap', None): 2.358e6,
                                    ('diffus_phase', 'Liq'): 1.493e-9}


@pytest.mark.component
class TestSeawaterPropertySolution_2(PropertyRegressionTest):
    def configure(self):
        self.prop_pack = props.SeawaterParameterBlock
        self.param_args = {}

        self.solver = 'ipopt'
        self.optarg = {'nlp_scaling_method': 'user-scaling'}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1,
                             ('flow_mass_phase_comp', ('Liq', 'TDS')): 1e2}
        self.state_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 0.95,
                           ('flow_mass_phase_comp', ('Liq', 'TDS')): 0.05,
                           ('temperature', None): 273.15 + 10,
                           ('pressure', None): 100e5}
        self.regression_solution = {('mass_frac_phase_comp', ('Liq', 'H2O')): 0.95,
                                    ('mass_frac_phase_comp', ('Liq', 'TDS')): 0.05,
                                    ('dens_mass_phase', 'Liq'): 1.039e3,
                                    ('dens_mass_solvent', None): 999.5,
                                    ('flow_vol_phase', 'Liq'): 9.628e-4,
                                    ('conc_mass_phase_comp', ('Liq', 'H2O')): 986.8,
                                    ('conc_mass_phase_comp', ('Liq', 'TDS')): 51.93,
                                    ('flow_mol_phase_comp', ('Liq', 'H2O')): 52.73,
                                    ('flow_mol_phase_comp', ('Liq', 'TDS')): 1.592,
                                    ('mole_frac_phase_comp', ('Liq', 'H2O')): 0.9707,
                                    ('mole_frac_phase_comp', ('Liq', 'TDS')): 2.931e-2,
                                    ('molality_comp', 'TDS'): 1.676,
                                    ('visc_d_phase', 'Liq'): 1.443e-3,
                                    ('osm_coeff', None): 0.9106,
                                    ('pressure_osm', None): 3.591e6,
                                    ('enth_mass_phase', 'Liq'): 3.897e4,
                                    ('pressure_sat', None): 1.194e3,
                                    ('cp_phase', 'Liq'): 3.916e3,
                                    ('therm_cond_phase', 'Liq'): 0.5854,
                                    ('dh_vap', None): 2.353e6,
                                    ('diffus_phase', 'Liq'): 1.471e-9}


@pytest.mark.component
class TestSeawaterCalculateState_1(PropertyCalculateStateTest):
    def configure(self):
        self.prop_pack = props.SeawaterParameterBlock
        self.param_args = {}

        self.solver = 'ipopt'
        self.optarg = {'nlp_scaling_method': 'user-scaling'}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1e-1,
                             ('flow_mass_phase_comp', ('Liq', 'TDS')): 1e1}
        self.var_args = {('flow_vol_phase', 'Liq'): 2e-2,
                         ('mass_frac_phase_comp', ('Liq', 'TDS')): 0.05,
                         ('temperature', None): 273.15 + 25,
                         ('pressure', None): 5e5}
        self.state_solution = {('flow_mass_phase_comp', ('Liq', 'H2O')): 19.66,
                               ('flow_mass_phase_comp', ('Liq', 'TDS')): 1.035}


@pytest.mark.component
class TestNaClCalculateState_2(PropertyCalculateStateTest):
    def configure(self):
        self.prop_pack = props.SeawaterParameterBlock
        self.param_args = {}

        self.solver = 'ipopt'
        self.optarg = {'nlp_scaling_method': 'user-scaling'}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1,
                             ('flow_mass_phase_comp', ('Liq', 'TDS')): 1e2}
        self.var_args = {('flow_vol_phase', 'Liq'): 1e-3,
                         ('pressure_osm', None): 100e5,
                         ('temperature', None): 273.15 + 25,
                         ('pressure', None): 5e5}
        self.state_solution = {('flow_mass_phase_comp', ('Liq', 'H2O')): 0.9604,
                               ('flow_mass_phase_comp', ('Liq', 'TDS')): 0.1231}

@pytest.mark.component
class TestNaClCalculateState_3(PropertyCalculateStateTest):
    def configure(self):
        self.prop_pack = props.SeawaterParameterBlock
        self.param_args = {}

        self.solver = 'ipopt'
        self.optarg = {'nlp_scaling_method': 'user-scaling'}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1,
                             ('flow_mass_phase_comp', ('Liq', 'TDS')): 1e2}
        self.var_args = {('flow_vol_phase', 'Liq'): 1e-3,
                         ('mass_frac_phase_comp', ('Liq', 'TDS')): 0.06,
                         ('pressure_sat', None): 3e4,
                         ('pressure', None): 5e5}
        self.state_solution = {('flow_mass_phase_comp', ('Liq', 'H2O')): 0.9605,
                               ('flow_mass_phase_comp', ('Liq', 'TDS')): 0.0613,
                               ('temperature', None): 343.05}




###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

import pytest

import pyomo.environ as pyo
from pyomo.util.check_units import assert_units_consistent
from pyomo.network import Arc, Port

from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.scaling import badly_scaled_var_generator
from idaes.generic_models.unit_models import Mixer, Separator, Product, Feed
from idaes.generic_models.unit_models.mixer import MomentumMixingType

from watertap.property_models.NaCl_prop_pack import NaClParameterBlock
from watertap.unit_models.pump_isothermal import Pump
from watertap.unit_models.reverse_osmosis_0D import ReverseOsmosis0D

from watertap.examples.flowsheets.lsrro.lsrro import (build, set_operating_conditions,
        initialize, optimize_set_up, solve, display_system, display_design, display_state)

class _TestLSRRO:

    @pytest.fixture(scope="class")
    def model(self):
        return build(number_of_stages=self.number_of_stages)

    @pytest.mark.unit
    def test_build(self, model):
        fs = model.fs

        # model set up
        assert isinstance(model, pyo.ConcreteModel)
        assert isinstance(fs, FlowsheetBlock)
        assert isinstance(fs.properties, NaClParameterBlock)
        assert isinstance(fs.costing, pyo.Block)

        # stages
        assert fs.NumberOfStages.value == self.number_of_stages
        assert len(fs.StageSet) == self.number_of_stages
        assert list(fs.StageSet) == list(range(1,self.number_of_stages+1))
        assert list(fs.LSRRO_StageSet) == list(range(2,self.number_of_stages+1))
        assert list(fs.NonFinal_StageSet) == list(range(1,self.number_of_stages))

        assert isinstance(fs.NumberOfStages, pyo.Param)
        assert isinstance(fs.StageSet, pyo.RangeSet)
        assert isinstance(fs.LSRRO_StageSet, pyo.RangeSet)
        assert isinstance(fs.NonFinal_StageSet, pyo.RangeSet)

        # units
        assert isinstance(fs.feed, Feed)
        assert len(fs.feed) == 1

        assert isinstance(fs.product, Product)
        assert len(fs.product) == 1
        assert isinstance(fs.disposal, Product)
        assert len(fs.disposal) == 1

        assert isinstance(fs.PrimaryPumps, Pump)
        assert len(fs.PrimaryPumps) == self.number_of_stages
        assert isinstance(fs.BoosterPumps, Pump)
        assert len(fs.BoosterPumps) == self.number_of_stages - 1
        assert isinstance(fs.EnergyRecoveryDevice, Pump)
        assert len(fs.EnergyRecoveryDevice) == 1

        assert isinstance(fs.ROUnits, ReverseOsmosis0D)
        assert len(fs.ROUnits) == self.number_of_stages

        # unit model options
        for mx in fs.Mixers.values():
            assert isinstance(mx.upstream, Port)
            assert isinstance(mx.downstream, Port)
            assert isinstance(mx.pressure_equality_constraints, pyo.Constraint)

        for ro in fs.ROUnits.values():
            assert isinstance(ro.deltaP, pyo.Var)

        # additional Pyomo elements
        assert isinstance(fs.water_recovery, pyo.Var)
        assert isinstance(fs.eq_water_recovery, pyo.Constraint)
        assert isinstance(fs.annual_water_production, pyo.Expression)
        assert isinstance(fs.specific_energy_consumption, pyo.Expression)

        # costing blocks and variables
        costing_units = ['PrimaryPumps', 'BoosterPumps', 'ROUnits', 'EnergyRecoveryDevice']
        for unit in costing_units:
            for blk in fs.component(unit).values():
                assert isinstance(blk.costing, pyo.Block)
                assert isinstance(blk.costing.capital_cost, pyo.Var)
                assert isinstance(blk.costing.operating_cost, pyo.Var)

        costing_var_names = ['capital_cost_total', 'investment_cost_total', 'operating_cost_MLC',
                             'operating_cost_total', 'LCOW']
        for varname in costing_var_names:
            assert isinstance(fs.costing.component(varname), pyo.Var)

        # arcs
        last_stage = self.number_of_stages
        arc_source_sink = [
                (fs.feed_to_pump, fs.feed.outlet, fs.PrimaryPumps[1].inlet),
                (fs.primary_RO_to_product, fs.ROUnits[1].permeate, fs.product.inlet),
                (fs.pump_to_stage, fs.PrimaryPumps[last_stage].outlet, fs.ROUnits[last_stage].inlet),
                (fs.stage_to_erd, fs.ROUnits[last_stage].retentate, fs.EnergyRecoveryDevice.inlet),
                (fs.erd_to_disposal, fs.EnergyRecoveryDevice.outlet, fs.disposal.inlet),
                ]

        for idx in range(1,self.number_of_stages):
            arc_source_sink.append((fs.pump_to_mixer[idx], fs.PrimaryPumps[idx].outlet, fs.Mixers[idx].upstream))
            arc_source_sink.append((fs.mixer_to_stage[idx], fs.Mixers[idx].outlet, fs.ROUnits[idx].inlet))
            arc_source_sink.append((fs.stage_to_pump[idx], fs.ROUnits[idx].retentate, fs.PrimaryPumps[idx+1].inlet))
        for idx in range(2,self.number_of_stages+1):
            arc_source_sink.append((fs.stage_to_eq_pump[idx], fs.ROUnits[idx].permeate, fs.BoosterPumps[idx].inlet))
            arc_source_sink.append((fs.eq_pump_to_mixer[idx], fs.BoosterPumps[idx].outlet, fs.Mixers[idx-1].downstream))

        for arc, src, dest in arc_source_sink:
            assert arc.src is src
            assert arc.dest is dest

        # additional bounds
        for blk in model.component_data_objects(pyo.Block, descend_into=True):
            # NaCl solubility limit
            if hasattr(blk, 'mass_frac_phase_comp'):
                blk.mass_frac_phase_comp['Liq', 'NaCl'].ub == 0.26

        # high-level checks
        model.compute_statistics()
        assert model.statistics.number_of_variables == self.number_of_variables
        assert model.statistics.number_of_constraints == self.number_of_constraints
        assert model.statistics.number_of_objectives == 1

        assert_units_consistent(fs)

    @staticmethod
    def _test_fixed_value(pyovar, val):
        assert pyovar.fixed
        assert pyovar.value == val

    @staticmethod
    def _test_no_badly_scaled_vars(m):
        for v,_ in badly_scaled_var_generator(m):
            raise Exception(f"Badly scaled variable {v.name}")

    @pytest.mark.component
    def test_set_operating_conditions(self, model):
        set_operating_conditions(model)

        fs = model.fs

        # feed
        self._test_fixed_value(fs.feed.pressure[0], 101325)
        self._test_fixed_value(fs.feed.temperature[0], 298.15)
        self._test_fixed_value(fs.feed.flow_mass_phase_comp[0, 'Liq', 'NaCl'], 70./1000.)
        self._test_fixed_value(fs.feed.flow_mass_phase_comp[0, 'Liq', 'H2O'], (1. - 70./1000.))

        # pumps
        for pump in fs.PrimaryPumps.values():
            self._test_fixed_value(pump.control_volume.properties_out[0].pressure, 75e5)
            self._test_fixed_value(pump.efficiency_pump[0], 0.75)
        for pump in fs.BoosterPumps.values():
            assert not pump.control_volume.properties_out[0].pressure.fixed
            self._test_fixed_value(pump.efficiency_pump[0], 0.75)
        for pump in fs.EnergyRecoveryDevice.values():
            self._test_fixed_value(pump.control_volume.properties_out[0].pressure, 101325)
            self._test_fixed_value(pump.efficiency_pump[0], 0.80)

        # RO units
        for idx, ro in fs.ROUnits.items():
            for acomp in ro.A_comp.values():
                self._test_fixed_value(acomp, 4.2e-12)
            for bcomp in ro.B_comp.values():
                if idx > 1:
                    self._test_fixed_value(bcomp, 3.5e-8*100.)
                else:
                    self._test_fixed_value(bcomp, 3.5e-8)

            self._test_fixed_value(ro.channel_height, 1e-3)
            self._test_fixed_value(ro.spacer_porosity, 0.97)
            self._test_fixed_value(ro.area, 100/float(idx))
            self._test_fixed_value(ro.width, 5)
            self._test_fixed_value(ro.permeate.pressure[0], 101325)

        assert degrees_of_freedom(model) == 0
        self._test_no_badly_scaled_vars(model)

    @pytest.mark.component
    def test_initialize(self, model, initialization_data):
        initialize(model)
        for var, val in initialization_data.items():
            assert pyo.value(var) == pytest.approx(val, rel=1e-5)
        self._test_no_badly_scaled_vars(model)

    @pytest.mark.component
    def test_simulation(self, model, simulation_data):
        solve(model)
        for var, val in simulation_data.items():
            assert pyo.value(var) == pytest.approx(val, rel=1e-5)
        self._test_no_badly_scaled_vars(model)

    @pytest.mark.component
    def test_display_system(self, model, capsys):
        display_system(model)
        captured = capsys.readouterr()
        assert captured.out == self.display_system

    @pytest.mark.component
    def test_display_design(self, model, capsys):
        display_design(model)
        captured = capsys.readouterr()
        assert captured.out == self.display_design

    @pytest.mark.component
    def test_display_state(self, model, capsys):
        display_state(model)
        captured = capsys.readouterr()
        assert captured.out == self.display_state

    @pytest.mark.component
    def test_optimize_set_up(self, model, optimization_data):
        optimize_set_up(model, water_recovery=optimization_data[model.fs.water_recovery])
        fs = model.fs

        for pump in fs.PrimaryPumps.values():
            assert not pump.control_volume.properties_out[0].pressure.fixed
        for pump in fs.BoosterPumps.values():
            assert not pump.control_volume.properties_out[0].pressure.fixed
        for idx, ro in fs.ROUnits.items():
            assert not ro.area.fixed
            assert not ro.width.fixed
            assert not ro.N_Re_io[0, 'in'].fixed
            if idx > 1:
                for bcomp in ro.B_comp.values():
                    assert not bcomp.fixed
            else:
                for bcomp in ro.B_comp.values():
                    assert bcomp.fixed

    @pytest.mark.component
    def test_optimize(self, model, optimization_data):
        solve(model)
        for var, val in optimization_data.items():
            assert pyo.value(var) == pytest.approx(val, rel=1e-5)


class TestLSRRO_1Stage(_TestLSRRO):

    number_of_stages = 1

    number_of_variables = 290
    number_of_constraints = 189

    display_system = \
"""----system metrics----
Feed: 1.00 kg/s, 70000 ppm
Product: 0.180 kg/s, 1592 ppm
Volumetric water recovery: 18.0%
Energy Consumption: 3.8 kWh/m3
Levelized cost of water: 2.48 $/m3
"""
    display_design = \
"""--decision variables--
Stage 1 operating pressure 75.0 bar
Stage 1 membrane area      100.0 m2
Stage 1 salt perm. coeff.  0.1 LMH
"""
    display_state = \
"""--------state---------
Feed                : 1.000 kg/s, 70000 ppm, 1.0 bar
Primary Pump 1 out  : 1.000 kg/s, 70000 ppm, 75.0 bar
RO 1 permeate       : 0.180 kg/s, 1592 ppm, 1.0 bar
RO 1 retentate      : 0.820 kg/s, 84977 ppm, 72.9 bar
Stage 1 Volumetric water recovery: 18.89%, Salt rejection: 97.84%
Disposal            : 0.820 kg/s, 84977 ppm, 1.0 bar
Product             : 0.180 kg/s, 1592 ppm, 1.0 bar
"""

    @pytest.fixture(scope="class")
    def initialization_data(self, model):
        data = pyo.ComponentMap()
        fs = model.fs

        data[fs.product.flow_mass_phase_comp[0,'Liq','H2O']]   = 0.179331
        data[fs.product.flow_mass_phase_comp[0,'Liq','NaCl']]  = 0.286037e-3
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','H2O']]  = 0.750668
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','NaCl']] = 0.697139e-1
        data[fs.costing.LCOW]   = 1.0
        data[fs.water_recovery] = 0.5

        return data

    @pytest.fixture(scope="class")
    def simulation_data(self, model):
        data = pyo.ComponentMap()
        fs = model.fs

        data[fs.product.flow_mass_phase_comp[0,'Liq','H2O']]   = 0.179331
        data[fs.product.flow_mass_phase_comp[0,'Liq','NaCl']]  = 0.286037e-3
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','H2O']]  = 0.750668
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','NaCl']] = 0.697139e-1
        data[fs.costing.LCOW]   = 2.48442
        data[fs.water_recovery] = 0.179618

        return data

    @pytest.fixture(scope="class")
    def optimization_data(self, model):
        data = pyo.ComponentMap()
        fs = model.fs

        data[fs.product.flow_mass_phase_comp[0,'Liq','H2O']]   = 0.385923
        data[fs.product.flow_mass_phase_comp[0,'Liq','NaCl']]  = 0.482338e-3
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','H2O']]  = 0.544077
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','NaCl']] = 0.695177e-1
        data[fs.costing.LCOW]   = 1.32260
        data[fs.water_recovery] = 0.386405

        return data


class TestLSRRO_2Stage(_TestLSRRO):

    number_of_stages = 2

    number_of_variables = 534
    number_of_constraints = 378

    display_system = \
"""----system metrics----
Feed: 1.00 kg/s, 70000 ppm
Product: 0.297 kg/s, 926 ppm
Volumetric water recovery: 29.7%
Energy Consumption: 6.3 kWh/m3
Levelized cost of water: 1.98 $/m3
"""
    display_design = \
"""--decision variables--
Stage 1 operating pressure 75.0 bar
Stage 1 membrane area      100.0 m2
Stage 1 salt perm. coeff.  0.1 LMH
Stage 2 operating pressure 75.0 bar
Stage 2 membrane area      50.0 m2
Stage 2 salt perm. coeff.  12.6 LMH
"""
    display_state = \
"""--------state---------
Feed                : 1.000 kg/s, 70000 ppm, 1.0 bar
Primary Pump 1 out  : 1.000 kg/s, 70000 ppm, 75.0 bar
Mixer 1 recycle     : 0.308 kg/s, 37757 ppm, 75.0 bar
Mixer 1 out         : 1.308 kg/s, 62417 ppm, 75.0 bar
RO 1 permeate       : 0.297 kg/s, 926 ppm, 1.0 bar
RO 1 retentate      : 1.011 kg/s, 80454 ppm, 72.0 bar
Stage 1 Volumetric water recovery: 23.74%, Salt rejection: 98.58%
Primary Pump 2 out  : 1.011 kg/s, 80454 ppm, 75.0 bar
RO 2 permeate       : 0.308 kg/s, 37757 ppm, 1.0 bar
RO 2 retentate      : 0.703 kg/s, 99118 ppm, 74.0 bar
Stage 2 Volumetric water recovery: 31.38%, Salt rejection: 54.50%
Booster Pump 2 out  : 0.308 kg/s, 37757 ppm, 75.0 bar
Disposal            : 0.703 kg/s, 99118 ppm, 1.0 bar
Product             : 0.297 kg/s, 926 ppm, 1.0 bar
"""

    @pytest.fixture(scope="class")
    def initialization_data(self, model):
        data = pyo.ComponentMap()
        fs = model.fs

        data[fs.product.flow_mass_phase_comp[0,'Liq','H2O']]   = 0.298029
        data[fs.product.flow_mass_phase_comp[0,'Liq','NaCl']]  = 0.274413e-3
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','H2O']]  = 0.633927
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','NaCl']] = 0.697161e-1
        data[fs.costing.LCOW]   = 1.0
        data[fs.water_recovery] = 0.5

        return data

    @pytest.fixture(scope="class")
    def simulation_data(self, model):
        data = pyo.ComponentMap()
        fs = model.fs

        data[fs.product.flow_mass_phase_comp[0,'Liq','H2O']]   = 0.296269
        data[fs.product.flow_mass_phase_comp[0,'Liq','NaCl']]  = 0.274578e-3
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','H2O']]  = 0.633730
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','NaCl']] = 0.697254e-1
        data[fs.costing.LCOW]   = 1.98397
        data[fs.water_recovery] = 0.296544

        return data

    @pytest.fixture(scope="class")
    def optimization_data(self, model):
        data = pyo.ComponentMap()
        fs = model.fs

        data[fs.product.flow_mass_phase_comp[0,'Liq','H2O']]   = 0.732053
        data[fs.product.flow_mass_phase_comp[0,'Liq','NaCl']]  = 0.451359e-3
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','H2O']]  = 0.197947
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','NaCl']] = 0.695486e-1
        data[fs.costing.LCOW]   = 1.21780
        data[fs.water_recovery] = 0.732504

        return data


class TestLSRRO_3Stage(_TestLSRRO):

    number_of_stages = 3

    number_of_variables = 778
    number_of_constraints = 567

    display_system = \
"""----system metrics----
Feed: 1.00 kg/s, 70000 ppm
Product: 0.330 kg/s, 823 ppm
Volumetric water recovery: 33.0%
Energy Consumption: 7.8 kWh/m3
Levelized cost of water: 2.08 $/m3
"""
    display_design = \
"""--decision variables--
Stage 1 operating pressure 75.0 bar
Stage 1 membrane area      100.0 m2
Stage 1 salt perm. coeff.  0.1 LMH
Stage 2 operating pressure 75.0 bar
Stage 2 membrane area      50.0 m2
Stage 2 salt perm. coeff.  12.6 LMH
Stage 3 operating pressure 75.0 bar
Stage 3 membrane area      33.3 m2
Stage 3 salt perm. coeff.  12.6 LMH
"""
    display_state = \
"""--------state---------
Feed                : 1.000 kg/s, 70000 ppm, 1.0 bar
Primary Pump 1 out  : 1.000 kg/s, 70000 ppm, 75.0 bar
Mixer 1 recycle     : 0.348 kg/s, 32663 ppm, 75.0 bar
Mixer 1 out         : 1.348 kg/s, 60351 ppm, 75.0 bar
RO 1 permeate       : 0.330 kg/s, 823 ppm, 1.0 bar
RO 1 retentate      : 1.019 kg/s, 79612 ppm, 71.9 bar
Stage 1 Volumetric water recovery: 25.55%, Salt rejection: 98.69%
Primary Pump 2 out  : 1.019 kg/s, 79612 ppm, 75.0 bar
Mixer 2 recycle     : 0.175 kg/s, 45135 ppm, 75.0 bar
Mixer 2 out         : 1.194 kg/s, 74561 ppm, 75.0 bar
RO 2 permeate       : 0.348 kg/s, 32663 ppm, 1.0 bar
RO 2 retentate      : 0.845 kg/s, 91835 ppm, 73.8 bar
Stage 2 Volumetric water recovery: 30.10%, Salt rejection: 57.51%
Booster Pump 2 out  : 0.348 kg/s, 32663 ppm, 75.0 bar
Primary Pump 3 out  : 0.845 kg/s, 91835 ppm, 75.0 bar
RO 3 permeate       : 0.175 kg/s, 45135 ppm, 1.0 bar
RO 3 retentate      : 0.670 kg/s, 104020 ppm, 74.4 bar
Stage 3 Volumetric water recovery: 21.40%, Salt rejection: 52.48%
Booster Pump 3 out  : 0.175 kg/s, 45135 ppm, 75.0 bar
Disposal            : 0.670 kg/s, 104020 ppm, 1.0 bar
Product             : 0.330 kg/s, 823 ppm, 1.0 bar
"""

    @pytest.fixture(scope="class")
    def initialization_data(self, model):
        data = pyo.ComponentMap()
        fs = model.fs

        data[fs.product.flow_mass_phase_comp[0,'Liq','H2O']]   = 0.351684
        data[fs.product.flow_mass_phase_comp[0,'Liq','NaCl']]  = 0.269335e-3
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','H2O']]  = 0.606601
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','NaCl']] = 0.695767e-1
        data[fs.costing.LCOW]   = 1.0
        data[fs.water_recovery] = 0.5

        return data

    @pytest.fixture(scope="class")
    def simulation_data(self, model):
        data = pyo.ComponentMap()
        fs = model.fs

        data[fs.product.flow_mass_phase_comp[0,'Liq','H2O']]   = 0.329390
        data[fs.product.flow_mass_phase_comp[0,'Liq','NaCl']]  = 0.271454e-3
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','H2O']]  = 0.600609
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','NaCl']] = 0.697285e-1
        data[fs.costing.LCOW]   = 2.07793
        data[fs.water_recovery] = 0.329661

        return data

    @pytest.fixture(scope="class")
    def optimization_data(self, model):
        data = pyo.ComponentMap()
        fs = model.fs

        data[fs.product.flow_mass_phase_comp[0,'Liq','H2O']]   = 0.732036
        data[fs.product.flow_mass_phase_comp[0,'Liq','NaCl']]  = 0.445064e-3
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','H2O']]  = 0.197964
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','NaCl']] = 0.695545e-1
        data[fs.costing.LCOW]   = 1.51258
        data[fs.water_recovery] = 0.732481

        return data


'''
class TestLSRRO_NStage(_TestLSRRO):

    number_of_stages = 

    number_of_variables = 
    number_of_constraints = 

    display_system = 
    display_design = 
    display_state = 

    @pytest.fixture(scope="class")
    def initialization_data(self, model):
        data = pyo.ComponentMap()
        fs = model.fs

        data[fs.product.flow_mass_phase_comp[0,'Liq','H2O']]   = 
        data[fs.product.flow_mass_phase_comp[0,'Liq','NaCl']]  = 
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','H2O']]  = 
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','NaCl']] = 
        data[fs.costing.LCOW]   = 
        data[fs.water_recovery] = 

        return data

    @pytest.fixture(scope="class")
    def simulation_data(self, model):
        data = pyo.ComponentMap()
        fs = model.fs

        data[fs.product.flow_mass_phase_comp[0,'Liq','H2O']]   = 
        data[fs.product.flow_mass_phase_comp[0,'Liq','NaCl']]  = 
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','H2O']]  = 
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','NaCl']] = 
        data[fs.costing.LCOW]   = 
        data[fs.water_recovery] = 

        return data

    @pytest.fixture(scope="class")
    def optimization_data(self, model):
        data = pyo.ComponentMap()
        fs = model.fs

        data[fs.product.flow_mass_phase_comp[0,'Liq','H2O']]   = 
        data[fs.product.flow_mass_phase_comp[0,'Liq','NaCl']]  = 
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','H2O']]  = 
        data[fs.disposal.flow_mass_phase_comp[0,'Liq','NaCl']] = 
        data[fs.costing.LCOW]   = 
        data[fs.water_recovery] = 

        return data
'''
